#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Test cases for the convert script used to access the DataConverter."""

import os
import logging
from setuptools import distutils
from click.testing import CliRunner
import pytest
import h5py
from pynxtools.nexus import nexus  # noqa: E402
import pynxtools.dataconverter.convert as dataconverter
from pynxtools.dataconverter.readers.base.reader import BaseReader


def move_xarray_file_to_tmp(tmp_path):
    """Moves the xarray file, which is used to test linking into the tmp_path directory."""
    test_file_path = os.path.join(os.path.dirname(__file__),
                                  "../data/dataconverter/readers/mpes")
    distutils.file_util.copy_file(os.path.join(test_file_path, "xarray_saved_small_calibration.h5"),
                                  os.path.join(tmp_path, "xarray_saved_small_calibration.h5"))


def restore_xarray_file_from_tmp(tmp_path):
    """Restores the xarray file from the tmp_path directory."""
    test_file_path = os.path.join(os.path.dirname(__file__),
                                  "../data/dataconverter/readers/mpes")
    os.remove(os.path.join(test_file_path, "xarray_saved_small_calibration.h5"))
    distutils.file_util.move_file(os.path.join(tmp_path, "xarray_saved_small_calibration.h5"),
                                  os.path.join(test_file_path, "xarray_saved_small_calibration.h5"))


@pytest.mark.parametrize("cli_inputs", [
    pytest.param([
        "--nxdl",
        "NXcontainer",
    ], id="exists-in-contributed"),
    pytest.param([
        "--nxdl",
        "NXarchive",
    ], id="exists-in-applications"),
    pytest.param([
        "--nxdl",
        "NXdoesnotexist",
    ], id="does-not-exist")
])
def test_find_nxdl(cli_inputs):
    """Unit test to check if dataconverter can find NXDLs in contributed/applications folder."""
    cli_inputs.extend(["--reader", "example"])

    runner = CliRunner()
    result = runner.invoke(dataconverter.convert_cli, cli_inputs)
    if "NXdoesnotexist" in cli_inputs:
        assert isinstance(result.exception, FileNotFoundError)
    else:
        assert isinstance(result.exception, Exception)
        assert "The chosen NXDL isn't supported by the selected reader." in str(result.exception)


def test_get_reader():
    """Unit test for the helper function to get a reader."""
    assert isinstance(dataconverter.get_reader("example")(), BaseReader)


def test_get_names_of_all_readers():
    """Unit test for the helper function to get all readers."""
    assert "example" in dataconverter.get_names_of_all_readers()


@pytest.mark.parametrize("cli_inputs", [
    pytest.param([
        "--nxdl",
        "NXtest",
        "--generate-template"
    ], id="generate-template"),
    pytest.param([], id="nxdl-not-provided"),
    pytest.param([
        "--nxdl",
        "NXtest",
        "--input-file",
        "test_input"
    ], id="input-file")
])
def test_cli(caplog, cli_inputs):
    """A test for the convert CLI."""
    runner = CliRunner()
    result = runner.invoke(dataconverter.convert_cli, cli_inputs)
    if "--generate-template" in cli_inputs:
        assert result.exit_code == 0
        assert "\"/ENTRY[entry]/NXODD_name/int_value\": \"None\"," in caplog.text
    elif "--input-file" in cli_inputs:
        assert "test_input" in caplog.text
    elif result.exit_code == 2:
        assert "Error: Missing option '--nxdl'" in result.output


def test_links_and_virtual_datasets(tmp_path):
    """A test for the convert CLI to check whether a Dataset object is created,

when  the template contains links."""
    move_xarray_file_to_tmp(tmp_path)

    dirpath = os.path.join(os.path.dirname(__file__),
                           "../data/dataconverter/readers/example")
    runner = CliRunner()
    result = runner.invoke(dataconverter.convert_cli, [
        "--nxdl",
        "NXtest",
        "--reader",
        "example",
        "--input-file",
        os.path.join(dirpath, "testdata.json"),
        "--output",
        os.path.join(tmp_path, "test_output.h5")
    ])

    assert result.exit_code == 0
    test_nxs = h5py.File(os.path.join(tmp_path, "test_output.h5"), "r")
    assert 'entry/test_link/internal_link' in test_nxs
    assert isinstance(test_nxs["entry/test_link/internal_link"], h5py.Dataset)
    assert 'entry/test_link/external_link' in test_nxs
    assert isinstance(test_nxs["entry/test_link/external_link"], h5py.Dataset)
    assert 'entry/test_virtual_dataset/concatenate_datasets' in test_nxs
    assert isinstance(test_nxs["entry/test_virtual_dataset/concatenate_datasets"], h5py.Dataset)
    assert 'entry/test_virtual_dataset/sliced_dataset' in test_nxs
    assert isinstance(test_nxs["entry/test_virtual_dataset/sliced_dataset"], h5py.Dataset)
    # pylint: disable=no-member
    assert test_nxs["entry/test_virtual_dataset/sliced_dataset"].shape == (10, 10, 5)
    assert 'entry/test_virtual_dataset/sliced_dataset2' in test_nxs
    assert isinstance(test_nxs["entry/test_virtual_dataset/sliced_dataset2"], h5py.Dataset)
    assert test_nxs["entry/test_virtual_dataset/sliced_dataset2"].shape == (10, 10, 10)
    assert 'entry/test_virtual_dataset/sliced_dataset3' in test_nxs
    assert isinstance(test_nxs["entry/test_virtual_dataset/sliced_dataset3"], h5py.Dataset)
    assert test_nxs["entry/test_virtual_dataset/sliced_dataset3"].shape == (10, 10, 10, 2)

    restore_xarray_file_from_tmp(tmp_path)


def test_compression(tmp_path):
    """A test for the convert CLI to check whether a Dataset object is compressed."""

    dirpath = os.path.join(os.path.dirname(__file__),
                           "../data/dataconverter/readers/example")

    move_xarray_file_to_tmp(tmp_path)

    dataconverter.convert(
        [os.path.join(dirpath, "testdata.json")],
        "example",
        "NXtest",
        os.path.join(tmp_path, "test_output.h5")
    )

    test_nxs = h5py.File(os.path.join(tmp_path, "test_output.h5"), "r")
    assert 'entry/test_compression/compressed_data' in test_nxs
    assert isinstance(test_nxs['/entry/test_compression/compressed_data'], h5py.Dataset)
    # pylint: disable=no-member
    assert test_nxs['/entry/test_compression/compressed_data'].compression == 'gzip'
    assert test_nxs['/entry/test_compression/not_to_compress'].compression is None

    restore_xarray_file_from_tmp(tmp_path)


def test_mpes_writing(tmp_path):
    """Check if mpes example can be reproduced"""
    # dataconverter
    dirpath = os.path.join(os.path.dirname(__file__), "../data/dataconverter/readers/mpes")
    dataconverter.convert((os.path.join(dirpath, "xarray_saved_small_calibration.h5"),
                           os.path.join(dirpath, "config_file.json")),
                          "mpes", "NXmpes",
                          os.path.join(tmp_path, "mpes.small_test.nxs"),
                          False, False)
    # check generated nexus file
    test_data = os.path.join(tmp_path, 'mpes.small_test.nxs')
    logger = logging.getLogger(__name__)
    logger.setLevel(logging.DEBUG)
    handler = logging. \
        FileHandler(os.path.join(tmp_path, 'mpes_test.log'), 'w')
    formatter = logging.Formatter('%(levelname)s - %(message)s')
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    nexus_helper = nexus.HandleNexus(logger, test_data, None, None)
    nexus_helper.process_nexus_master_file(None)
    with open(os.path.join(tmp_path, 'mpes_test.log'), 'r', encoding='utf-8') as logfile:
        log = logfile.readlines()
    with open(
        os.path.join(dirpath, 'Ref_nexus_mpes.log'),
        'r',
        encoding='utf-8'
    ) as logfile:
        ref_log = logfile.readlines()
    assert log == ref_log


def test_eln_data(tmp_path):
    """Check if the subsections in the eln_data.yml file work."""
    dirpath = os.path.join(os.path.dirname(__file__), "../data/dataconverter/readers/mpes")
    dataconverter.convert((os.path.join(dirpath, "xarray_saved_small_calibration.h5"),
                           os.path.join(dirpath, "config_file.json"),
                           os.path.join(dirpath, "eln_data.yaml")),
                          "mpes", "NXmpes",
                          os.path.join(tmp_path, "mpes.small_test.nxs"),
                          False, False)


def test_eln_data_subsections(tmp_path):
    """Check if the subsections in the eln_data.yml file work."""
    dirpath = os.path.join(os.path.dirname(__file__), "../data/dataconverter/readers/json_yml")
    dataconverter.convert((os.path.join(dirpath, "eln_data_w_subsections.yaml",),),
                          "hall", "NXroot",
                          os.path.join(tmp_path, "hall.nxs"),
                          False, False)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Parse metadata in generic (like reported by an OASIS ELN) schema."""

# pylint: disable=no-member,duplicate-code

import numpy as np

import flatdict as fd

import yaml

# from ase.data import chemical_symbols

from pynxtools.dataconverter.readers.em_om.utils.versioning \
    import NX_EM_OM_ADEF_NAME, NX_EM_OM_ADEF_VERSION
from pynxtools.dataconverter.readers.em_om.utils.versioning \
    import NX_EM_OM_EXEC_NAME, NX_EM_OM_EXEC_VERSION

from pynxtools.dataconverter.readers.em_om.utils.handed_cartesian \
    import REFERENCE_FRAMES, AXIS_DIRECTIONS, is_cs_well_defined

# example how to check against different types of Euler angle conventions
# from pynxtools.dataconverter.readers.em_om.utils.euler_angle_convention \
#    import which_euler_convention

# example how to check if set of conventions matches to some suggestion in the literature
# from pynxtools.dataconverter.readers.em_om.utils.msmse_convention \
#     import is_consistent_with_msmse_convention


class NxEmOmGenericElnSchemaParser:
    """Parse eln_data.yaml dump file content generated from an (e.g. OASIS) ELN.

    """

    def __init__(self, file_name: str, entry_id: int, pattern_simulation: bool):
        """Fill template with ELN pieces of information."""
        self.pattern_simulation = pattern_simulation
        print(f"Extracting data from ELN file: {file_name}")
        if (file_name.rsplit('/', 1)[-1].startswith("eln_data")
                or file_name.startswith("eln_data")) and entry_id > 0:
            self.entry_id = entry_id
            self.file_name = file_name
            with open(self.file_name, "r", encoding="utf-8") as stream:
                self.yml = fd.FlatDict(yaml.safe_load(stream), delimiter=":")
        else:
            self.entry_id = 1
            self.file_name = ""
            self.yml = {}

        # if "ElectronBackscatterDiffraction" in self.yml:
        #     self.yml = self.yml["ElectronBackscatterDiffraction"]

    def parse(self, template: dict) -> dict:
        """Extract metadata from generic ELN text file to respective NeXus objects."""
        print("Parsing ELN input...")
        print(f"{self.file_name}")
        print(f"{self.entry_id}")
        self.parse_entry_section(template)
        self.parse_user_section(template)
        if self.pattern_simulation is False:
            self.parse_commerical_on_the_fly_section(template)
            self.parse_measurement_section(template)
            self.parse_calibration_section(template)
        self.parse_rotation_convention_section(template)
        self.parse_processing_frame_section(template)
        self.parse_sample_frame_section(template)
        self.parse_detector_frame_section(template)
        self.parse_gnomonic_projection_section(template)
        if self.pattern_simulation is False:
            self.parse_indexing(template)
        return template

    def parse_entry_section(self, template: dict) -> dict:
        """"Parse entry section."""
        print("Parse entry...")
        src = "entry"
        trg = f"/ENTRY[entry{self.entry_id}]/"
        if (self.yml[f"{src}:attr_version"] == NX_EM_OM_ADEF_VERSION) \
                and (self.yml[f"{src}:definition"] == NX_EM_OM_ADEF_NAME):
            template[f"{trg}@version"] = NX_EM_OM_ADEF_VERSION
            template[f"{trg}definition"] = NX_EM_OM_ADEF_NAME
            template[f"{trg}PROGRAM[program1]/program"] = NX_EM_OM_EXEC_NAME
            template[f"{trg}PROGRAM[program1]/program/@version"] = NX_EM_OM_EXEC_VERSION
        if ("program" in self.yml[src].keys()) \
                and ("program__attr_version" in self.yml[src].keys()):
            template[f"{trg}PROGRAM[program2]/program"] \
                = self.yml[f"{src}:program"]
            template[f"{trg}PROGRAM[program2]/program/@version"] \
                = self.yml[f"{src}:program__attr_version"]
        # check that versions NX_EM_OM_* match
        req_field_names = ["definition", "start_time", "end_time",
                           "workflow_description", "workflow_identifier"]
        for field in req_field_names:
            if field in self.yml[src].keys():
                template[f"{trg}{field}"] = self.yml[f"{src}:{field}"]

        return template

    def parse_user_section(self, template: dict) -> dict:
        """Parse user section."""
        print("Parse user...")
        src = "user"
        if not isinstance(self.yml[src], list):
            return template
        user_id = 1
        for user_list in self.yml[src]:
            trg = f"/ENTRY[entry{self.entry_id}]/USER[user{user_id}]/"
            field_names = [
                "name", "email", "affiliation", "address",
                "orcid", "orcid_platform",
                "telephone_number", "role",
                "social_media_name", "social_media_platform"]
            for field_name in field_names:
                if field_name in user_list.keys():
                    template[f"{trg}{field_name}"] = user_list[field_name]
            user_id += 1
        return template

    def parse_commerical_on_the_fly_section(self, template: dict) -> dict:
        """Parse the on-the-fly section."""
        print("Parse commercial on-the-fly")
        src = "commercial_on_the_fly_indexing"
        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing/on_the_fly_indexing/"
        if ("program" in self.yml[src].keys()) \
                and ("program__attr_version" in self.yml[src].keys()):
            template[f"{trg}PROGRAM[program1]/program"] \
                = self.yml[f"{src}:program"]
            template[f"{trg}PROGRAM[program1]/program/@version"] \
                = self.yml[f"{src}:program__attr_version"]
        if "results_file" in self.yml[src].keys():
            template[f"{trg}origin"] = self.yml[f"{src}:results_file"]
        if "results_file__attr_version" in self.yml[src].keys():
            template[f"{trg}origin/@version"] \
                = self.yml[f"{src}:results_file__attr_version"]
        template[f"{trg}path"] = str("undefined")
        # NEW ISSUE: this is a bug not results_file version in eln but path !!
        return template

    def parse_measurement_section(self, template: dict) -> dict:
        """Parse measurement section establishing link between NXem_ebsd and NXem."""
        print("Parse measurement...")
        src = "measurement"
        field_names = ["origin", "path"]
        trg = f"/ENTRY[entry{self.entry_id}]/experiment/acquisition/"
        for field_name in field_names:
            if field_name in self.yml[src].keys():
                template[f"{trg}{field_name}"] = self.yml[f"{src}:{field_name}"]
        if "origin__attr_version" in self.yml[src].keys():
            template[f"{trg}origin/@version"] = self.yml[f"{src}:origin__attr_version"]
        return template

    def parse_calibration_section(self, template: dict) -> dict:
        """Parse calibration section."""
        print("Parse calibration...")
        src = "calibration"
        field_names = ["origin", "path"]
        trg = f"/ENTRY[entry{self.entry_id}]/experiment/calibration/"
        for field_name in field_names:
            if field_name in self.yml[src].keys():
                template[f"{trg}{field_name}"] = self.yml[f"{src}:{field_name}"]
        if "origin__attr_version" in self.yml[src].keys():
            template[f"{trg}origin/@version"] = self.yml[f"{src}:origin__attr_version"]
        return template

    def parse_rotation_convention_section(self, template: dict) -> dict:
        """Parse rotation conventions."""
        src = "rotation_conventions"
        trg = f"/ENTRY[entry{self.entry_id}]/conventions/rotation_conventions/"
        terms = [
            "three_dimensional_rotation_handedness",
            "rotation_convention",
            "euler_angle_convention",
            "axis_angle_convention"]
        for term in terms:
            if term in self.yml[src].keys():
                template[f"{trg}{term}"] = self.yml[f"{src}:{term}"].lower()
        # one term named differently in ELN than in NeXus appdef template keyword
        if "sign_convention" in self.yml[src].keys():
            template[f"{trg}orientation_parameterization_sign_convention"] \
                = self.yml[f"{src}:sign_convention"].lower()
        # if desired one could check conventions are consistent with specific ones
        return template

    def parse_processing_frame_section(self, template: dict) -> dict:
        """Parse processing reference frame definitions."""
        axes_names = ["x", "y", "z"]
        src = "processing_reference_frame"
        trg = f"/ENTRY[entry{self.entry_id}]/conventions/processing_reference_frame/"
        if "reference_frame_type" in self.yml[src].keys():
            if self.yml[f"{src}:reference_frame_type"] in REFERENCE_FRAMES:
                template[f"{trg}reference_frame_type"] \
                    = self.yml[f"{src}:reference_frame_type"]
            xyz_directions = ["undefined", "undefined", "undefined"]
            xyz_aliases = ["", "", ""]
            for idx in np.arange(0, 3):
                axis_name = axes_names[idx]
                if f"{axis_name}axis_direction" in self.yml[src].keys():
                    if self.yml[f"{src}:{axis_name}axis_direction"] in AXIS_DIRECTIONS:
                        xyz_directions[idx] = self.yml[f"{src}:{axis_name}axis_direction"]
                if f"{axis_name}axis_alias" in self.yml[src].keys():
                    xyz_aliases[idx] = self.yml[f"{src}:{axis_name}axis_alias"]

            if is_cs_well_defined(self.yml[f"{src}:reference_frame_type"], xyz_directions):
                for idx in np.arange(0, 3):
                    axis_name = axes_names[idx]
                    template[f"{trg}{axis_name}axis_direction"] = xyz_directions[idx]
                    template[f"{trg}{axis_name}axis_alias"] = xyz_aliases[idx]
            if "origin" in self.yml[src].keys():
                template[f"{trg}origin"] = self.yml[f"{src}:origin"]
        return template

    def parse_sample_frame_section(self, template: dict) -> dict:
        """Parse sample reference frame definitions."""
        axes_names = ["x", "y", "z"]
        src = "sample_reference_frame"
        trg = f"/ENTRY[entry{self.entry_id}]/conventions/sample_reference_frame/"
        if "reference_frame_type" in self.yml[src].keys():
            if self.yml[f"{src}:reference_frame_type"] in REFERENCE_FRAMES:
                template[f"{trg}reference_frame_type"] \
                    = self.yml[f"{src}:reference_frame_type"]
            xyz_directions = ["undefined", "undefined", "undefined"]
            for idx in np.arange(0, 3):
                axis_name = axes_names[idx]
                if f"{axis_name}axis_direction" in self.yml[src].keys():
                    if self.yml[f"{src}:{axis_name}axis_direction"] in AXIS_DIRECTIONS:
                        xyz_directions[idx] = self.yml[f"{src}:{axis_name}axis_direction"]
            if is_cs_well_defined(self.yml[f"{src}:reference_frame_type"], xyz_directions):
                for idx in np.arange(0, 3):
                    axis_name = axes_names[idx]
                    template[f"{trg}{axis_name}axis_direction"] = xyz_directions[idx]
            if "origin" in self.yml[src].keys():
                template[f"{trg}origin"] = self.yml[f"{src}:origin"]
        return template

    def parse_detector_frame_section(self, template: dict) -> dict:
        """Parse detector reference frame definitions."""
        axes_names = ["x", "y", "z"]
        src = "detector_reference_frame"
        trg = f"/ENTRY[entry{self.entry_id}]/conventions/detector_reference_frame/"
        if "reference_frame_type" in self.yml[src].keys():
            if self.yml[f"{src}:reference_frame_type"] in REFERENCE_FRAMES:
                template[f"{trg}reference_frame_type"] \
                    = self.yml[f"{src}:reference_frame_type"]
            xyz_directions = ["undefined", "undefined", "undefined"]
            for idx in np.arange(0, 3):
                axis_name = axes_names[idx]
                if f"{axis_name}axis_direction" in self.yml[src].keys():
                    if self.yml[f"{src}:{axis_name}axis_direction"] in AXIS_DIRECTIONS:
                        xyz_directions[idx] = self.yml[f"{src}:{axis_name}axis_direction"]
            if is_cs_well_defined(self.yml[f"{src}:reference_frame_type"], xyz_directions):
                for idx in np.arange(0, 3):
                    axis_name = axes_names[idx]
                    template[f"{trg}{axis_name}axis_direction"] = xyz_directions[idx]
            if "origin" in self.yml[src].keys():
                template[f"{trg}origin"] = self.yml[f"{src}:origin"]
        # ... <-- is the same except for the src name
        return template

    def parse_gnomonic_projection_section(self, template: dict) -> dict:
        """Parse for the gnomonic projection."""
        axes_names = ["x", "y", "z"]
        src = "gnomonic_projection:gnomonic_projection_reference_frame"
        trg = f"/ENTRY[entry{self.entry_id}]/conventions" \
              f"/gnomonic_projection_reference_frame/"
        if "reference_frame_type" in self.yml[src].keys():
            if self.yml[f"{src}:reference_frame_type"] in REFERENCE_FRAMES:
                template[f"{trg}reference_frame_type"] \
                    = self.yml[f"{src}:reference_frame_type"]
            xyz_directions = ["undefined", "undefined", "undefined"]
            for idx in np.arange(0, 3):
                axis_name = axes_names[idx]
                if f"{axis_name}axis_direction" in self.yml[src].keys():
                    if self.yml[f"{src}:{axis_name}axis_direction"] in AXIS_DIRECTIONS:
                        xyz_directions[idx] = self.yml[f"{src}:{axis_name}axis_direction"]
            if is_cs_well_defined(self.yml[f"{src}:reference_frame_type"], xyz_directions):
                for idx in np.arange(0, 3):
                    axis_name = axes_names[idx]
                    template[f"{trg}{axis_name}axis_direction"] = xyz_directions[idx]
            if "origin" in self.yml[src].keys():
                template[f"{trg}origin"] = self.yml[f"{src}:origin"]

        src = "gnomonic_projection:pattern_centre"
        trg = f"/ENTRY[entry{self.entry_id}]/conventions/pattern_centre/"
        axes_names = ["x", "y"]
        field_names = ["axis_boundary_convention", "axis_normalization_direction"]
        for idx in np.arange(0, 2):
            axis_name = axes_names[idx]
            for field_name in field_names:
                if f"{axis_name}{field_name}" in self.yml[src].keys():
                    template[f"{trg}{axis_name}{field_name}"] \
                        = self.yml[f"{src}:{axis_name}{field_name}"]
        return template

    def parse_indexing(self, template: dict) -> dict:
        """Parse indexing section."""
        print("Parse indexing...")
        src = "indexing"
        trg = f"/ENTRY[entry{self.entry_id}]/experiment/indexing"
        if "method" in self.yml[src].keys():
            template[f"{trg}/method"] = self.yml[f"{src}:method"]
        return template



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utility functions for generation of atom probe NeXus datasets for dev purposes."""

# pylint: disable=no-member,duplicate-code,too-many-instance-attributes,attribute-defined-outside-init

# MK::2023/05/04 the code in this file can currently not be used when
# developers have an environment which uses ase==3.19.0 and numpy>=1.2x

import hashlib

import datetime

from typing import List

import numpy as np

import ase
from ase.lattice.cubic import FaceCenteredCubic
from ase.data import atomic_numbers, atomic_masses, chemical_symbols

from ifes_apt_tc_data_modeling.utils.utils \
    import create_isotope_vector, isotope_vector_to_nuclid_list, \
    isotope_vector_to_human_readable_name, MAX_NUMBER_OF_ATOMS_PER_ION, MQ_EPSILON

# do not use ase directly any longer for NIST isotopes, instead this syntatic equivalent
# from ifes_apt_tc_data_modeling.utils.nist_isotope_data \
#     import isotopes

from pynxtools.dataconverter.readers.apm.utils.apm_versioning \
    import NX_APM_ADEF_NAME, NX_APM_ADEF_VERSION, NX_APM_EXEC_NAME, NX_APM_EXEC_VERSION

from pynxtools.dataconverter.readers.apm.utils.apm_load_ranging \
    import add_unknown_iontype


# parameter affecting reconstructed positions and size
CRYSTAL_ORIENTATION = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
# MK::add analysis how large aggregate has to be
RECON_SIZE = (50, 50, 300)
RECON_ATOM_SPACING = 5.
RECON_HEIGHT = 300.  # angstroem
RECON_RADIUS = 50.  # angstroem
MAX_COMPONENTS = 5  # how many different molecular ions in one dataset/entry
MAX_ATOMS = 10  # determine power-law fraction of n_atoms per ion
MULTIPLES_FACTOR = 0.6  # controls how likely multiple ions are synthesized
# the higher this factor the more uniformly and more likely multiplicity > 1
MAX_CHARGE_STATE = 4  # highest allowed charge_state of an ion
MAX_ATOMIC_NUMBER = 94  # do not include heavier atoms than Plutonium
MAX_USERS = 4


class ApmCreateExampleData:
    """A synthesized dataset meant to be used for development purposes only!."""

    def __init__(self, synthesis_id):
        # assure deterministic behaviour of the PRNG
        np.random.seed(seed=synthesis_id)

        self.n_entries = 1
        print("Generating one random example NXem entry...")
        self.entry_id = 1
        # reconstructed dataset and mass-to-charge state ratio values
        # like what is traditionally available via the POS file format
        self.xyz: List[float] = []
        self.m_z: List[float] = []

        # synthesizing realistic datasets for atom probe tomography
        # would require a physical model of the field evaporation process,
        # a model of the detector, and the reconstruction and ranging algorithms.
        # A cutting edge view of such simulations is described here
        # https://arxiv.org/abs/2209.05997, this demands very costly HPC computing
        # As this goes beyond what is numerically feasible, strong assumptions are made:
        # we are interested in having positions on a grid (real specimen geometry)
        #     details do not matter, positions are assumed exact without noise
        # reflect that signal comes from elementary and molecular ions
        #     details do not matter but sample from the entire periodic table
        # we wish to have a histogram of mass-to-charge-state ratio values
        # this is called a mass spectrum in the field of atom probe microscopy
        #     no tails in peaks are modelled as this depends on the pulser, physics
        #     and is in fact not yet well understood physically for general materials

    def create_reconstructed_positions(self):
        """Create aggregate of specifically oriented

        crystallographic unit cells and take the positions
        in this lattice to carve out a portion of the
        point cloud representing the reconstruction positions."""
        # https://wiki.fysik.dtu.dk/ase/ase/lattice.html#module-ase.lattice
        # ase length units in angstroem!
        # https://wiki.fysik.dtu.dk/ase/ase/units.html#units

        # assumptions:
        # identity orientation, no periodic boundary conditions
        print(f"Using the following version of ase {ase.__version__}")
        xyz = np.asarray(FaceCenteredCubic(directions=CRYSTAL_ORIENTATION,
                                           size=RECON_SIZE, symbol="Cu",
                                           latticeconstant=RECON_ATOM_SPACING,
                                           pbc=(0, 0, 0)).get_positions(), np.float32)
        # Cu will be ignored, only the lattice with positions is relevant
        centre_of_mass = np.asarray([np.mean(xyz[:, 0]),
                                     np.mean(xyz[:, 1]),
                                     np.mean(xyz[:, 2])], np.float32)
        # print("Centre of mass of ASE lattice is (with coordinates in angstroem)")
        # print(centre_of_mass)
        xyz = xyz - centre_of_mass
        centre_of_mass = np.asarray([np.mean(xyz[:, 0]),
                                     np.mean(xyz[:, 1]),
                                     np.mean(xyz[:, 2])], np.float32)
        # print("Updated centre of mass")
        # print(centre_of_mass)
        # axis_aligned_bbox = np.asarray([np.min(xyz[:, 0]), np.max(xyz[:, 0]),
        #                                 np.min(xyz[:, 1]), np.max(xyz[:, 1]),
        #                                 np.min(xyz[:, 2]), np.max(xyz[:, 2])])
        # displace origin
        origin = centre_of_mass
        # print("Building a cylinder of radius " + str(RECON_RADIUS * 0.1) + " nm"
        #       + " and height " + str(RECON_HEIGHT * 0.1) + " nm")
        mask = None
        mask = xyz[:, 2] <= (origin[2] + 0.5 * RECON_HEIGHT)
        mask &= xyz[:, 2] >= (origin[2] - 0.5 * RECON_HEIGHT)
        mask &= ((xyz[:, 0] - origin[0])**2
                 + (xyz[:, 1] - origin[1])**2) <= RECON_RADIUS**2
        self.xyz = xyz[mask]
        shift = [0., 0., 0.5 * RECON_HEIGHT]
        for idx in np.arange(0, 3):
            self.xyz[:, idx] += shift[idx]
        self.xyz *= 0.1  # from angstroem to nm
        print("Created a geometry for a reconstructed dataset, shape is")
        print(np.shape(self.xyz))
        # self.aabb3d = np.asarray([np.min(self.xyz[:, 0]), np.max(self.xyz[:, 0]),
        #                           np.min(self.xyz[:, 1]), np.max(self.xyz[:, 1]),
        #                           np.min(self.xyz[:, 2]), np.max(self.xyz[:, 2])])

    def place_atoms_from_periodic_table(self):
        """Sample elements from the periodic table

        create (hypothetical) charged molecular ions from them
        and evaluate their mass-to-charge-state ratio to be used
        as values in the example dataset."""

        # uniform random model for how many different ions
        # !! warning: for real world datasets this depends on real specimen composition
        self.n_components = int(np.random.uniform(low=1, high=MAX_COMPONENTS))
        # print("Number of ions in the composition is " + str(self.n_components))

        # power law model for multiplicity of molecular ions
        # !! warning: for real world datasets depends on evaporation physics
        self.n_ivec = np.asarray(np.linspace(1,
                                             MAX_ATOMS,
                                             num=MAX_ATOMS,
                                             endpoint=True), np.float64)
        accept_reject = MULTIPLES_FACTOR**self.n_ivec
        accept_reject = np.cumsum(accept_reject) / np.sum(accept_reject)
        unifrnd = np.random.uniform(low=0., high=1., size=(self.n_components,))
        self.multiplicity = np.ones((self.n_components,))
        for idx in np.arange(0, len(accept_reject) - 1):
            mask = unifrnd[:] >= accept_reject[idx]
            mask &= unifrnd[:] < accept_reject[idx + 1]
            self.multiplicity[mask] = self.n_ivec[idx]
        self.multiplicity = np.asarray(self.multiplicity, np.uint32)

        # uniform model for distribution of charge states
        # !! warning: for real world datasets actual ion charge depends
        # on (evaporation) physics, very complicated in fact a topic of current research
        self.charge_state = np.asarray(np.random.uniform(low=1,
                                                         high=MAX_CHARGE_STATE,
                                                         size=(self.n_components,)),
                                       np.uint32)

        # compose for each component randomly sampled hypothetical molecular ions
        # uniform random model which elements to pick from periodic table of elements
        # !! warning: for real world datasets molecular ions found research dependant
        # !! often research in many groups is strongly focused on specific
        # materials and abundance, toxic nature of some elements forbids
        # experiments with these, like Plutonium or, also reason for synthetic data
        value_to_pse_symbol_lookup = {}
        for key, val in atomic_numbers.items():
            if key != "X":
                value_to_pse_symbol_lookup[val] = key

        composition = []  # list of tuples, one for each composition
        for idx in np.arange(0, self.n_components):
            isotope_vector = []
            mass_sum = 0.
            # sample atoms for building the ion
            sampled_elements = np.asarray(
                np.random.uniform(low=1, high=MAX_ATOMIC_NUMBER,
                                  size=(self.multiplicity[idx],)), np.uint32)

            for val in sampled_elements:
                symbol = value_to_pse_symbol_lookup[val]
                isotope_vector.append(symbol)
                mass_sum += atomic_masses[atomic_numbers[symbol]]

            composition.append((isotope_vector,
                                self.charge_state[idx],
                                mass_sum / self.charge_state[idx],
                                np.float64(np.random.uniform(low=1, high=100))))

        weighting_factor_sum = 0.
        for idx in np.arange(0, self.n_components):
            weighting_factor_sum += composition[idx][3]

        # normalize all compositions
        # (weighting_factor_sum)
        self.nrm_composition = []
        # print(composition)
        for idx in np.arange(0, self.n_components):
            self.nrm_composition.append((
                composition[idx][0],
                composition[idx][1],
                composition[idx][2],
                composition[idx][3] / weighting_factor_sum))

        self.nrm_composition.sort(key=lambda a: a[3])  # sort asc. for composition
        accept_reject = [0.]
        for idx in self.nrm_composition:
            accept_reject.append(idx[3])
        accept_reject = np.cumsum(accept_reject)
        assert self.xyz != [], \
            "self.xyz must not be an empty dataset, create a geometry first!"
        # print("Accept/reject sampling m/q values for "
        #       + str(np.shape(self.xyz)[0]) + " ions")

        unifrnd = np.random.uniform(low=0., high=1., size=(np.shape(self.xyz)[0],))
        self.m_z = np.empty((np.shape(self.xyz)[0],))
        self.m_z[:] = np.nan
        for idx in np.arange(0, len(accept_reject) - 1):
            mask = unifrnd[:] >= accept_reject[idx]
            mask &= unifrnd[:] < accept_reject[idx + 1]
            self.m_z[mask] = self.nrm_composition[idx][2]
            # print(self.nrm_composition[idx])
            # print(np.sum(mask) / np.shape(self.xyz)[0])
        # print(np.shape(self.m_z))
        # assert np.sum(self.m_z == np.nan) == 0, "Not all m/q values defined!"

    def composition_to_ranging_definitions(self, template: dict) -> dict:
        """Create ranging definitions based on composition."""
        assert len(self.nrm_composition) > 0, "Composition is not defined!"
        trg = f"/ENTRY[entry{self.entry_id}]/atom_probe/ranging/"
        template[f"{trg}PROGRAM[program1]/program"] = NX_APM_EXEC_NAME
        template[f"{trg}PROGRAM[program1]/program/@version"] = NX_APM_EXEC_VERSION

        trg = f"/ENTRY[entry{self.entry_id}]/atom_probe/ranging/peak_identification/"
        template[f"{trg}PROGRAM[program1]/program"] = "synthetic"
        template[f"{trg}PROGRAM[program1]/program/@version"] = "synthetic data"

        add_unknown_iontype(template, self.entry_id)

        ion_id = 1
        for tpl in self.nrm_composition:
            path = f"{trg}ION[ion{ion_id}]/"
            ivec = create_isotope_vector(tpl[0])
            template[f"{path}isotope_vector"] \
                = np.reshape(np.asarray(ivec, np.uint16),
                             (1, MAX_NUMBER_OF_ATOMS_PER_ION))
            # template[path + "isotope_vector/@units"] = ""
            template[f"{path}charge_state"] = np.int8(tpl[1])
            template[f"{path}mass_to_charge_range"] = np.reshape(
                np.asarray([tpl[2], tpl[2] + MQ_EPSILON], np.float32), (1, 2))
            template[f"{path}mass_to_charge_range/@units"] = "Da"
            nuclid_list = np.zeros([2, 32], np.uint16)
            nuclid_list = isotope_vector_to_nuclid_list(ivec)
            template[f"{path}nuclid_list"] = np.asarray(nuclid_list, np.uint16)
            template[path + "name"] = isotope_vector_to_human_readable_name(ivec, tpl[1])
            ion_id += 1

        trg = f"/ENTRY[entry{self.entry_id}]/atom_probe/ranging/"
        template[f"{trg}number_of_ion_types"] = np.uint32(ion_id)
        template[f"{trg}maximum_number_of_atoms_per_molecular_ion"] = np.uint32(32)

        return template

    def emulate_entry(self, template: dict) -> dict:
        """Copy data in entry section."""
        # check if required fields exists and are valid
        # print("Parsing entry...")
        trg = f"/ENTRY[entry{self.entry_id}]/"
        template[f"{trg}definition"] = NX_APM_ADEF_NAME
        template[f"{trg}@version"] = NX_APM_ADEF_VERSION
        template[f"{trg}PROGRAM[program1]/program"] = NX_APM_EXEC_NAME
        template[f"{trg}PROGRAM[program1]/program/@version"] = NX_APM_EXEC_VERSION
        template[f"{trg}start_time"] = datetime.datetime.now().astimezone().isoformat()
        template[f"{trg}end_time"] = datetime.datetime.now().astimezone().isoformat()
        msg = '''
              WARNING: These are mocked data !!
              They are meant to be used exclusively
              for verifying NOMAD search capabilities.
              '''
        template[f"{trg}experiment_description"] = msg
        experiment_identifier \
            = str(f"R{np.random.choice(100, 1)[0]}-{np.random.choice(100000, 1)[0]}")
        template[f"{trg}experiment_identifier"] = experiment_identifier
        template[f"{trg}run_number"] = experiment_identifier.split("-")[1]
        template[f"{trg}operation_mode"] \
            = str(np.random.choice(["apt", "fim", "apt_fim"], 1)[0])
        return template

    def emulate_user(self, template: dict) -> dict:
        """Copy data in user section."""
        # check if required fields exists and are valid
        # print("Parsing user...")
        prefix = f"/ENTRY[entry{self.entry_id}]/"
        user_names = np.unique(
            np.random.choice(["Sherjeel", "MarkusK", "Dierk", "Baptiste",
                              "Alexander", "Lorenz", "Sophie", "Stefan",
                              "Katharina", "Florian", "Daniel", "Sandor",
                              "Carola", "Andrea", "Hampus", "Pepe", "Lauri",
                              "MarkusS", "Christoph", "Claudia"],
                             1 + np.random.choice(MAX_USERS, 1)))
        user_id = 1
        for name in user_names:
            trg = f"{prefix}USER[user{user_id}]/"
            template[f"{trg}name"] = str(name)
            user_id += 1
        return template

    def emulate_specimen(self, template: dict) -> dict:
        """Copy data in specimen section."""
        # check if required fields exists and are valid
        # print("Parsing specimen...")
        trg = f"/ENTRY[entry{self.entry_id}]/specimen/"
        assert len(self.nrm_composition) > 0, "Composition list is empty!"
        unique_elements = set()
        for tpl in self.nrm_composition:
            symbol_lst = tpl[0]
            for symbol in symbol_lst:
                assert isinstance(symbol, str), "symbol is not a string!"
                if (symbol in chemical_symbols) & (symbol != "X"):
                    unique_elements.add(str(symbol))
        print(f"Unique elements are: {list(unique_elements)}")
        template[f"{trg}atom_types"] = ", ".join(list(unique_elements))

        specimen_name = str(f"Mocked atom probe specimen {np.random.choice(1000, 1)[0]}")
        template[f"{trg}name"] = specimen_name
        template[f"{trg}sample_history"] = "n/a"
        template[f"{trg}preparation_date"] \
            = datetime.datetime.now().astimezone().isoformat()
        template[f"{trg}short_title"] \
            = specimen_name.replace("Mocked atom probe specimen ", "")
        template[f"{trg}description"] = "n/a"
        return template

    def emulate_control_software(self, template: dict) -> dict:
        """Copy data in control software section."""
        # print("Parsing control software...")
        trg = f"/ENTRY[entry{self.entry_id}]/atom_probe/control_software/"
        template[f"{trg}PROGRAM[program1]/program"] = "IVAS"
        template[f"{trg}PROGRAM[program1]/program/@version"] \
            = str(f"3.{np.random.choice(9, 1)[0]}.{np.random.choice(9, 1)[0]}")
        return template

    def emulate_instrument_header(self, template: dict) -> dict:
        """Copy data in instrument_header section."""
        # check if required fields exists and are valid
        # print("Parsing instrument header...")
        trg = f"/ENTRY[entry{self.entry_id}]/atom_probe/"
        template[f"{trg}instrument_name"] \
            = str(f"test instrument {np.random.choice(100, 1)[0]}")
        template[f"{trg}flight_path_length"] \
            = np.float64(np.random.normal(loc=1.0, scale=0.05))
        template[f"{trg}flight_path_length/@units"] = "m"
        return template

    def emulate_fabrication(self, template: dict) -> dict:
        """Copy data in fabrication section."""
        # print("Parsing fabrication...")
        trg = f"/ENTRY[entry{self.entry_id}]/atom_probe/FABRICATION[fabrication]/"
        template[f"{trg}vendor"] \
            = str(np.random.choice(["AMETEK/Cameca", "customized"], 1)[0])
        template[f"{trg}model"] \
            = str(np.random.choice(["LEAP3000", "LEAP4000", "LEAP5000",
                  "LEAP6000", "OxCart", "MTAP", "FIM"], 1)[0])
        template[f"{trg}identifier"] \
            = str(hashlib.sha256("IVAS".encode("utf-8")).hexdigest())
        # template[f"{trg}capabilities"] = ""
        return template

    def emulate_analysis_chamber(self, template: dict) -> dict:
        """Copy data in analysis_chamber section."""
        # print("Parsing analysis chamber...")
        trg = f"/ENTRY[entry{self.entry_id}]/atom_probe/analysis_chamber/"
        template[f"{trg}pressure"] = np.float64(
            np.random.normal(loc=1.0e-10, scale=0.2e-11))
        template[f"{trg}pressure/@units"] = "torr"
        return template

    def emulate_reflectron(self, template: dict) -> dict:
        """Copy data in reflectron section."""
        # print("Parsing reflectron...")
        trg = f"/ENTRY[entry{self.entry_id}]/atom_probe/REFLECTRON[reflectron]/"
        template[f"{trg}applied"] = bool(np.random.choice([0, 1], 1)[0])
        return template

    def emulate_local_electrode(self, template: dict) -> dict:
        """Copy data in local_electrode section."""
        # print("Parsing local electrode...")
        trg = f"/ENTRY[entry{self.entry_id}]/atom_probe/local_electrode/"
        template[f"{trg}name"] \
            = str(f"electrode {np.random.choice(1000, 1)[0]}")
        return template

    def emulate_detector(self, template: dict) -> dict:
        """Copy data in ion_detector section."""
        # print("Parsing detector...")
        trg = f"/ENTRY[entry{self.entry_id}]/atom_probe/ion_detector/"
        detector_model_type = str(np.random.choice(["cameca", "mcp", "custom"], 1)[0])
        template[f"{trg}type"] = detector_model_type
        template[f"{trg}name"] = detector_model_type
        template[f"{trg}model"] = detector_model_type
        template[f"{trg}serial_number"] = hashlib.sha256(
            detector_model_type.encode("utf-8")).hexdigest()
        return template

    def emulate_stage_lab(self, template: dict) -> dict:
        """Copy data in stage lab section."""
        # print("Parsing stage lab...")
        trg = f"/ENTRY[entry{self.entry_id}]/atom_probe/stage_lab/"
        template[f"{trg}base_temperature"] = np.float64(10 + np.random.choice(50, 1)[0])
        template[f"{trg}base_temperature/@units"] = "K"
        return template

    def emulate_specimen_monitoring(self, template: dict) -> dict:
        """Copy data in specimen_monitoring section."""
        # print("Parsing specimen monitoring...")
        trg = f"/ENTRY[entry{self.entry_id}]/atom_probe/specimen_monitoring/"
        eta = np.min((np.random.normal(loc=0.6, scale=0.1), 1.))
        template[f"{trg}detection_rate"] = np.float64(eta)
        template[f"{trg}initial_radius"] = np.float64(RECON_RADIUS * 0.1)
        template[f"{trg}initial_radius/@units"] = "nm"
        template[f"{trg}shank_angle"] = np.float64(0.)  # = np.random.choice(10, 1)[0]
        template[f"{trg}shank_angle/@units"] = "degree"
        return template

    def emulate_pulser(self, template: dict) -> dict:
        """Copy data in pulser section."""
        # print("Parsing pulser...")
        trg = f"/ENTRY[entry{self.entry_id}]/atom_probe/pulser/"
        pulse_mode = str(np.random.choice(
            ["laser", "voltage", "laser_and_voltage"], 1)[0])
        template[f"{trg}pulse_mode"] = pulse_mode
        template[f"{trg}pulse_fraction"] \
            = np.float64(np.random.normal(loc=0.1, scale=0.02))
        template[f"{trg}pulse_frequency"] \
            = np.float64(np.random.normal(loc=250, scale=10))
        template[f"{trg}pulse_frequency/@units"] = "kHz"
        if pulse_mode != "voltage":
            trg = f"/ENTRY[entry{self.entry_id}]/atom_probe/pulser/SOURCE[laser_source1]/"
            template[f"{trg}name"] = "laser"
            template[f"{trg}wavelength"] \
                = np.float64((30 + np.random.choice(30, 1)) * 1.0e-8)
            template[f"{trg}wavelength/@units"] = "m"
            template[f"{trg}pulse_energy"] \
                = np.float64(np.random.normal(loc=1.2e-11, scale=0.2e-12))
            template[f"{trg}pulse_energy/@units"] = "J"
            template[f"{trg}power"] \
                = np.float64(np.random.normal(loc=2.0e-8, scale=0.2e-9))
            template[f"{trg}power/@units"] = "W"
        return template

    def emulate_reconstruction(self, template: dict) -> dict:
        """Copy data in reconstruction section."""
        # print("Parsing reconstruction...")
        trg = f"/ENTRY[entry{self.entry_id}]/atom_probe/reconstruction/"
        src = f"/ENTRY[entry{self.entry_id}]/atom_probe/control_software/"
        template[f"{trg}PROGRAM[program1]/program"] \
            = template[f"{src}PROGRAM[program1]/program"]
        template[f"{trg}PROGRAM[program1]/program/@version"] \
            = template[f"{src}PROGRAM[program1]/program/@version"]
        template[f"{trg}protocol_name"] \
            = str(np.random.choice(["bas", "geiser", "gault", "cameca", "other"], 1)[0])
        template[f"{trg}parameter"] = "n/a"
        template[f"{trg}crystallographic_calibration"] = "n/a"
        return template

    def emulate_ranging(self, template: dict) -> dict:
        """Copy data in ranging section."""
        # print("Parsing ranging...")
        trg = f"/ENTRY[entry{self.entry_id}]/atom_probe/ranging/"
        src = f"/ENTRY[entry{self.entry_id}]/atom_probe/control_software/"
        template[f"{trg}PROGRAM[program1]/program"] \
            = template[f"{src}PROGRAM[program1]/program"]
        template[f"{trg}PROGRAM[program1]/program/@version"] \
            = template[f"{src}PROGRAM[program1]/program/@version"]
        return template

    def emulate_random_input_from_eln(self, template: dict) -> dict:
        """Emulate random input as could come from an ELN."""
        self.emulate_entry(template)
        self.emulate_user(template)
        self.emulate_specimen(template)

        self.emulate_control_software(template)
        # above call will set program and program/@version used later

        self.emulate_instrument_header(template)
        self.emulate_fabrication(template)
        self.emulate_analysis_chamber(template)
        self.emulate_reflectron(template)
        self.emulate_local_electrode(template)
        self.emulate_detector(template)
        self.emulate_stage_lab(template)
        self.emulate_specimen_monitoring(template)

        self.emulate_pulser(template)
        self.emulate_reconstruction(template)
        self.emulate_ranging(template)

        template[f"/ENTRY[entry{self.entry_id}]/atom_probe/status"] = "success"

        return template

    def synthesize(self, template: dict) -> dict:
        """Hand-over instantiated dataset to dataconverter template."""
        # heavy data, synthetic/mocked dataset
        for entry_id in np.arange(1, self.n_entries + 1):
            self.entry_id = entry_id
            print(f"Generating entry {self.entry_id}...")

            self.xyz = []
            self.m_z = []
            self.create_reconstructed_positions()
            self.place_atoms_from_periodic_table()
            self.composition_to_ranging_definitions(template)

            # metadata
            self.emulate_random_input_from_eln(template)

            # heavy numerical data, here the synthesized "measurement" data
            prefix = f"/ENTRY[entry{self.entry_id}]/atom_probe/"
            trg = f"{prefix}reconstruction/"
            template[f"{trg}reconstructed_positions"] \
                = {"compress": np.asarray(self.xyz, np.float32), "strength": 1}
            template[f"{trg}reconstructed_positions/@units"] = "nm"

            trg = f"{prefix}mass_to_charge_conversion/"
            template[f"{trg}mass_to_charge"] \
                = {"compress": np.asarray(self.m_z, np.float32), "strength": 1}
            template[f"{trg}mass_to_charge/@units"] = "Da"

        return template



import os
from pathlib import Path
from typing import Dict

import lxml.etree

from ..globals import directories
from ..globals.nxdl import NXDL_NAMESPACE
from ..nxdl import iter_definitions


def nxdl_indices() -> Dict[str, dict]:
    """For each directory under the NXDL root, create the content of an
    index file which link all NeXus class definition doc files.
    """
    indentation = " " * 4
    namespaces = {"nx": NXDL_NAMESPACE}
    sections = dict()
    root = directories.get_nxdl_root()

    for nxdl_file in iter_definitions():
        source = os.path.relpath(nxdl_file.parent, root)
        section = sections.get(source)
        if section is None:
            preample = _DIRNAME_TO_PREAMBLE.get(source)
            if preample is None:
                continue
            classes = []
            rst_lines = []
            index_file = directories.nxclass_build_root(nxdl_file) / "index.rst"
            sections[source] = {
                "index_file": index_file,
                "rst_lines": rst_lines,
                "classes": classes,
            }
            rst_lines.append(".. do NOT edit this file\n")
            rst_lines.append(f"   automatically generated by {__name__}\n")
            rst_lines.append("\n")
            rst_lines.append(preample)
            rst_lines.append("\n")
        else:
            classes = sections[source]["classes"]
            rst_lines = sections[source]["rst_lines"]

        nxclass_name = nxdl_file.with_suffix("").stem
        classes.append(nxclass_name)
        summary = get_nxclass_description(nxdl_file, namespaces)
        if "NXcg" in nxclass_name or "NXapm" in nxclass_name or "NXms" in nxclass_name:
            continue
        rst_lines.append("\n")
        rst_lines.append(f":ref:`{nxclass_name}`\n")
        rst_lines.append(f"{indentation}{summary}\n")

    # Create a table of content for each index file
    for section in sections.values():
        classes = section.pop("classes")
        rst_lines = section["rst_lines"]
        rst_lines.append("\n")
        rst_lines.append(".. toctree::\n")
        rst_lines.append(f"{indentation}:hidden:\n")
        rst_lines.append("\n")
        if "index_file" in section.keys():
            file = section["index_file"]
            print("----------", file)
            file = os.path.abspath(file)
        else:
            file = ""
            print("---------++++++++-", section)
        if file.endswith("contributed_definitions/index.rst"):
            rst_lines.append(f"{indentation}em-structure\n")
            rst_lines.append(f"{indentation}ellipsometry-structure\n")
            rst_lines.append(f"{indentation}mpes-structure\n")
            rst_lines.append(f"{indentation}apm-structure\n")
            rst_lines.append(f"{indentation}transport-structure\n")
            rst_lines.append(f"{indentation}sts-structure\n")
            rst_lines.append(f"{indentation}cgms-structure\n")
            rst_lines.append(f"{indentation}icme-structure\n")
            rst_lines.append(f"{indentation}sample-prep-structure\n")

        for cname in sorted(classes):
            rst_lines.append(f"{indentation}{cname}\n")

    return sections


def get_nxclass_description(nxdl_file: Path, namespaces) -> str:
    """
    get the summary line from each NXDL definition doc

    That's the first physical line of the doc string.
    """
    tree = lxml.etree.parse(nxdl_file)
    root = tree.getroot()
    nodes = root.xpath("nx:doc", namespaces=namespaces)
    if len(nodes) != 1:
        raise RuntimeError(f"wrong number of <doc> nodes in NXDL: {nxdl_file}")
    return nodes[0].text.strip().splitlines()[0]


_DIRNAME_TO_PREAMBLE = {
    "base_classes": """
.. index::
     ! see: class definitions; base class
     ! base class

.. _base.class.definitions:

Base Class Definitions
######################

A description of each NeXus base class definition is given.
NeXus base class definitions define the set of terms that
*might* be used in an instance of that class.
Consider the base classes as a set of *components*
that are used to construct a data file.
    """,
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    "applications": """
.. index::
     ! see: class definitions; application definition
     ! application definition

.. _application.definitions:

Application Definitions
#########################

A description of each NeXus application definition is given.
NeXus application definitions define the *minimum*
set of terms that
*must* be used in an instance of that class.
Application definitions also may define terms that
are optional in the NeXus data file.  The definition, in this case,
reserves the exact term by declaring its spelling and description.
Consider an application definition as a *contract*
between a data provider (such as the beam line control system) and a
data consumer (such as a data analysis program for a scientific technique)
that describes the information is certain to be available in a data file.

Use NeXus links liberally in data files to reduce duplication of data.
In application definitions involving raw data,
write the raw data in the :ref:`NXinstrument` tree and then link to it
from the location(s) defined in the relevant application definition.
    """,
    # - - - - - - - - - - - - - - - - - - - - - - - - - - - -
    "contributed_definitions": """
.. index::
     ! see: class definitions; contributed definition
     ! contributed definition

.. _contributed.definitions:

Contributed Definitions
#########################

A description of each NeXus contributed definition is given.
NXDL files in the NeXus contributed definitions include propositions from
the community for NeXus base classes or application definitions, as well
as other NXDL files for long-term archival by NeXus.  Consider the contributed
definitions as either in *incubation* or a special
case not for general use.  The :ref:`NIAC` is charged to review any new contributed
definitions and provide feedback to the authors before ratification
and acceptance as either a base class or application definition.

Some contributions are grouped together:
  :ref:`Optical Spectroscopy <Ellipsometry-Structure>`

  :ref:`Multi-dimensional Photoemission Spectroscopy <Mpes-Structure>`

  :ref:`Atomprobe Microscopy <Apm-Structure>`

  :ref:`Electron Microscopy <Em-Structure>`

  :ref:`Transport Measurements <Transport-Structure>`

  :ref:`Geometry and Microstructures  <CgmsFeatures-Structure>`


and others are simply listed here:

    """,
}



"""A generic reader for loading XPS (X-ray Photoelectron Spectroscopy) data
 file into mpes nxdl (NeXus Definition Language) template.
"""
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
from pathlib import Path
from typing import Any, Dict, Set, List
from typing import Tuple
import sys
import json

import yaml
import numpy as np

from pynxtools.dataconverter.readers.base.reader import BaseReader
from pynxtools.dataconverter.readers.xps.reader_utils import XpsDataFileParser
from pynxtools.dataconverter.readers.utils import flatten_and_replace, FlattenSettings
from pynxtools.dataconverter.template import Template
from pynxtools.dataconverter.helpers import extract_atom_types

np.set_printoptions(threshold=sys.maxsize)

XPS_TOCKEN = "@xps_tocken:"
XPS_DATA_TOCKEN = "@data:"
XPS_DETECTOR_TOCKEN = "@detector_data:"
ELN_TOCKEN = "@eln"
# Track entries for using for eln data
ENTRY_SET: Set[str] = set()
DETECTOR_SET: Set[str] = set()
POSSIBLE_ENTRY_PATH: Dict = {}


CONVERT_DICT = {
    'unit': '@units',
    'version': '@version',
    'User': 'USER[user]',
    'Instrument': 'INSTRUMENT[instrument]',
    'Source': 'SOURCE[source]',
    'Beam': 'BEAM[beam]',
    'Analyser': 'ELECTRONANALYSER[electronanalyser]',
    'Collectioncolumn': 'COLLECTIONCOLUMN[collectioncolumn]',
    'Energydispersion': 'ENERGYDISPERSION[energydispersion]',
    'Detector': 'DETECTOR[detector]',
    'Manipulator': 'MANIPULATOR[manipulator]',
    'Sample': 'SAMPLE[sample]',
    'Data': 'DATA[data]',
}

REPLACE_NESTED: Dict[str, str] = {}


def construct_entry_name(key):
    """Construct entry name from vendor, sample_name and region name"""

    components = key.split("/")
    try:
        # entry: vendor__sample__name_of_scan_rerion
        entry_name = (f'{components[2]}'
                      f'__'
                      f'{components[3].split("_", 1)[1]}'
                      f'__'
                      f'{components[5].split("_", 1)[1]}'
                      )
    except IndexError:
        entry_name = ""
    return entry_name


def find_entry_and_value(xps_data_dict,
                         key_part,
                         dt_typ):

    """Construct the entry name and pick up the corresponding data for
       that for that entry.
    """

    entries_values = {}
    if dt_typ == XPS_TOCKEN:

        for key, val in xps_data_dict.items():
            if key_part in key:
                entry = construct_entry_name(key)
                entries_values[entry] = val

    elif dt_typ in (XPS_DATA_TOCKEN, XPS_DETECTOR_TOCKEN):
        # entries_values = entry:{cycle0_scan0_chan0:xr.data}
        entries_values = xps_data_dict["data"]

    return entries_values


# pylint: disable=too-many-locals
# pylint: disable=too-many-statements
def fill_data_group(key,
                    entries_values,
                    config_dict,
                    template,
                    entry_set):
    """Fill out fileds and attributes for NXdata"""

    survey_count_ = 0
    count = 0

    for entry, xr_data in entries_values.items():
        entry_set.add(entry)
        modified_key = key.replace("entry", entry)
        modified_key = modified_key.replace("[data]/data", "[data]")
        root = key[0]
        modifid_entry = key[0:13]
        modifid_entry = modifid_entry.replace("entry", entry)

        template[f"{modifid_entry}/@default"] = "data"

        # Set first Survey as default for .nxs file
        if "Survey" in entry and survey_count_ == 0:
            survey_count_ = survey_count_ + 1
            template[f"{root}@default"] = entry

        # If no Survey set any scan for default
        if survey_count_ == 0 and count == 0:
            count = count + 1
            template[f"{root}@default"] = entry

        binding_energy_coord = None
        chan_count = "_chan"
        for data_var in xr_data.data_vars:
            scan = data_var
            # Collecting only accumulated counts
            # indivisual channeltron counts goes to detector data section
            if chan_count not in scan:
                key_indv_scn_dt = f"{modified_key}/{scan}"

                key_indv_scn_dt_unit = f"{key_indv_scn_dt}/@units"

                coord_nm = list(xr_data[data_var].coords)[0]
                binding_energy_coord = np.array(xr_data[data_var][coord_nm])
                template[key_indv_scn_dt] = xr_data[data_var].data
                template[key_indv_scn_dt_unit] = config_dict[f"{key}/@units"]

        key_data = f"{modified_key}/data"
        key_data_unit = f"{key_data}/@units"

        key_signal = f"{modified_key}/@signal"

        be_nm = "BE"
        be_index = 0
        key_be = f"{modified_key}/{be_nm}"
        key_be_unit = f"{key_be}/@units"
        key_be_axes = f"{modified_key}/@axes"
        key_be_ind = f"{modified_key}/@{be_nm}_indices"

        # setting up AXISNAME
        axisname = "AXISNAME[axisname]"
        long_name = "Binding Energy"
        key_ax_mn = f"{modified_key}/{axisname}"
        key_ax_ln_nm = f"{modified_key}/{axisname}/@long_name"

        key_nxclass = f"{modified_key}/@NX_class"

        template[key_signal] = "data"
        template[key_data] = np.mean([xr_data[x_arr].data
                                     for x_arr in xr_data.data_vars
                                     if "_chan" not in x_arr],
                                     axis=0)

        template[f"{key_data}_errors"] = \
            np.std([xr_data[x_arr].data
                    for x_arr in xr_data.data_vars
                    if "_chan" not in x_arr], axis=0)

        template[key_data_unit] = config_dict[f"{key}/@units"]
        template[key_be_unit] = "eV"
        template[key_be] = binding_energy_coord
        template[key_be_axes] = be_nm
        template[key_be_ind] = be_index
        template[key_nxclass] = "NXdata"
        template[key_ax_ln_nm] = long_name
        template[key_ax_mn] = be_index


def fill_detector_group(key,
                        entries_values,
                        config_dict,
                        template,
                        entry_set):
    """Fill out fileds and attributes for NXdetector/NXdata"""

    for entry, xr_data in entries_values.items():
        entry_set.add(entry)

        chan_count = "_chan"

        # Iteration over scan
        for data_var in xr_data.data_vars:

            if chan_count in data_var:
                detector_num = data_var.split("_chan")[-1]
                detector_nm = f"detector{detector_num}"
                DETECTOR_SET.add(detector_nm)
                scan_num = data_var.split("_scan")[-1].split("_chan")[0]
                scan_nm = f"scan_{scan_num}"
                modified_key = key.replace("entry", entry)
                modified_key = modified_key.replace("[detector]", f"[{detector_nm}]")
                modified_key = modified_key.replace("[data]", f"[{scan_nm}]")
                modified_key_unit = modified_key + "/@units"

                template[modified_key] = xr_data[data_var].data
                key_indv_chan_sginal = modified_key.replace("/raw", "/@signal")
                template[key_indv_chan_sginal] = "raw"
                template[modified_key_unit] = config_dict[f"{key}/@units"]


def fill_template_with_xps_data(config_dict,
                                xps_data_dict,
                                template,
                                entry_set):
    """Collect the xps data from xps_data_dict
        and store them into template. We use searching_keys
        for separating the data from xps_data_dict.
    """
    for key, value in config_dict.items():
        if XPS_DATA_TOCKEN in value:
            key_part = value.split(XPS_DATA_TOCKEN)[-1]
            entries_values = find_entry_and_value(xps_data_dict,
                                                  key_part,
                                                  dt_typ=XPS_DATA_TOCKEN)

            fill_data_group(key, entries_values, config_dict, template, entry_set)

        if XPS_DETECTOR_TOCKEN in value:
            key_part = value.split(XPS_DATA_TOCKEN)[-1]
            entries_values = find_entry_and_value(xps_data_dict,
                                                  key_part,
                                                  dt_typ=XPS_DETECTOR_TOCKEN)

            fill_detector_group(key, entries_values, config_dict, template, entry_set)

        if XPS_TOCKEN in value:
            tocken = value.split(XPS_TOCKEN)[-1]
            entries_values = find_entry_and_value(xps_data_dict,
                                                  tocken,
                                                  dt_typ=XPS_TOCKEN)
            for entry, ent_value in entries_values.items():
                entry_set.add(entry)
                modified_key = key.replace("[entry]", f"[{entry}]")
                template[modified_key] = ent_value
                try:
                    template[f"{modified_key}/@units"] = config_dict[f"{key}/@units"]
                except KeyError:
                    pass


# pylint: disable=too-many-branches
def fill_template_with_eln_data(eln_data_dict,
                                config_dict,
                                template,
                                entry_set):
    """Fill the template from provided eln data"""

    def fill_atom_types(key):
        atom_types: List = []
        field_value = eln_data_dict[key]

        if "chemical_formula" in key:
            atom_types = list(extract_atom_types(field_value))

        if field_value is None:
            return

        for entry in entry_set:
            modified_key = key.replace("[entry]", f"[{entry}]")
            template[modified_key] = field_value
            if atom_types:
                modified_key = modified_key.replace('chemical_formula', 'atom_types')
                template[modified_key] = ', '.join(atom_types)

    def fill_from_value(key):
        field_value = eln_data_dict[key]
        if not field_value:
            return
        # Do for all entry name
        for entry in entry_set:
            modified_key = key.replace("[entry]", f"[{entry}]")
            # Do for all detector
            if "[detector]" in key:
                for detector in DETECTOR_SET:
                    detr_key = modified_key.replace("[detector]", f"[{detector}]")
                    template[detr_key] = field_value
            else:
                template[modified_key] = field_value

    for key, val in config_dict.items():
        if ELN_TOCKEN in val:
            fill_atom_types(key)
        elif key in list(eln_data_dict.keys()):
            fill_from_value(key)


def concatenate_values(value1, value2):
    """
    Concatenate two values of same type to be stored
    in xps_data_dict. Dicts are merged and every other object is
    appended to a list.

    """
    if (isinstance(value1, dict) and isinstance(value2, dict)):
        concatenated = {**value1, **value2}
    else:
        if not isinstance(value1, list):
            value1 = [value1]
        if not isinstance(value2, list):
            value2 = [value2]
        concatenated = value1 + value2

    return concatenated


# pylint: disable=too-few-public-methods
class XPSReader(BaseReader):
    """ Reader for XPS.
    """

    supported_nxdls = [
        "NXmpes",
        # "NXmpes_xps"
    ]

    __config_files: Dict = {
        "xml": "config_file_xml.json",
        "sle": "config_file_xml.json",
        "txt": "config_file_scienta_txt.json",
    }

    def read(self,
             template: dict = None,
             file_paths: Tuple[str] = None,
             objects: Tuple[Any] = None,
             **kwargs) -> dict:
        """Reads data from given file and returns
        a filled template dictionary"""

        reader_dir = Path(__file__).parent

        xps_data_dict: Dict[str, Any] = {}
        eln_data_dict: Dict[str, Any] = {}

        for file in file_paths:
            file_ext = os.path.splitext(file)[1]

            if file_ext in [".yaml", ".yml"]:
                with open(file, mode="r", encoding="utf-8") as eln:
                    eln_data_dict = flatten_and_replace(
                        FlattenSettings(
                            yaml.safe_load(eln),
                            CONVERT_DICT,
                            REPLACE_NESTED
                        )
                    )
            elif file_ext in [".sle", ".xml", ".txt"]:
                data_dict = XpsDataFileParser([file]).get_dict(**kwargs)

                # If there are multiple input data files, make sure
                # that existing keys are not overwritten.
                existing = [
                    (key, xps_data_dict[key], data_dict[key]) for
                    key in set(xps_data_dict).intersection(data_dict)
                ]

                xps_data_dict = {**xps_data_dict, **data_dict}
                for (key, value1, value2) in existing:
                    xps_data_dict[key] = concatenate_values(
                        value1, value2)

                config_file = reader_dir.joinpath(
                    XPSReader.__config_files[file_ext.rsplit('.')[1]]
                )

            # This code is not very robust.
            elif file_ext == ".json":
                if "config_file" in file:
                    config_file = Path(file)

        with open(config_file, encoding="utf-8", mode="r") as cfile:
            config_dict = json.load(cfile)

        fill_template_with_xps_data(config_dict,
                                    xps_data_dict,
                                    template,
                                    ENTRY_SET)
        if eln_data_dict:
            fill_template_with_eln_data(eln_data_dict,
                                        config_dict,
                                        template,
                                        ENTRY_SET)
        else:
            raise ValueError("Eln file must be submited with some required fields and attributes.")

        final_template = Template()
        for key, val in template.items():
            if ("/ENTRY[entry]" not in key) and (val is not None):
                final_template[key] = val

        return final_template


READER = XPSReader



import textwrap
from pathlib import Path
from typing import List

import lxml.etree

from ..globals import directories
from ..globals.errors import NXDLParseError
from ..globals.nxdl import XSD_NAMESPACE
from ..utils.types import PathLike


class XSDDocGenerator:
    """
    Read the NXDL field types specification and find
    all the valid data types.  Write a restructured
    text (.rst) document for use in the NeXus manual in
    the NXDL chapter.
    """

    TITLE_MARKERS = "- + ~ ^ * @".split()  # used for underscoring section titles
    INDENTATION = " " * 4
    DATATYPE_DICT = {
        "basicComponent": """/xs:schema//xs:complexType[@name='basicComponent']""",
        "validItemName": """/xs:schema//xs:simpleType[@name='validItemName']""",
        "validNXClassName": """/xs:schema//xs:simpleType[@name='validNXClassName']""",
        "validTargetName": """/xs:schema//xs:simpleType[@name='validTargetName']""",
        "nonNegativeUnbounded": """/xs:schema//xs:simpleType[@name='nonNegativeUnbounded']""",
    }

    def __init__(self) -> None:
        self.ns = {"xs": XSD_NAMESPACE}
        self._rst_lines = None

    def __call__(self, xsd_file: PathLike) -> List[str]:
        self._rst_lines = list()
        xsd_file = Path(xsd_file)
        try:
            self._parse_xsd_file(xsd_file)
        except Exception:
            raise NXDLParseError(xsd_file)
        return self._rst_lines

    def _parse_xsd_file(self, xsd_file: Path):
        tree = lxml.etree.parse(str(xsd_file))

        self._print(f".. auto-generated by {__name__} -- DO NOT EDIT")
        self._print(ELEMENT_PREAMBLE)

        for name in sorted(ELEMENT_DICT):
            self._print("")
            self._print(".. index:: ! %s (NXDL element)\n" % name)
            self._print(".. _%s:\n" % name)
            self.print_title(name, indentLevel=0)
            self._print("\n")
            self._print(ELEMENT_DICT[name])
            self._print("\n")
            self.add_figure(name, indentLevel=0)

        self._print(DATATYPE_PREAMBLE)

        path_list = (
            "/xs:schema/xs:complexType[@name='attributeType']",
            "/xs:schema/xs:element[@name='definition']",
            "/xs:schema/xs:complexType[@name='definitionType']",
            "/xs:schema/xs:simpleType[@name='definitionTypeAttr']",
            "/xs:schema/xs:complexType[@name='dimensionsType']",
            "/xs:schema/xs:complexType[@name='docType']",
            "/xs:schema/xs:complexType[@name='enumerationType']",
            "/xs:schema/xs:complexType[@name='fieldType']",
            "/xs:schema/xs:complexType[@name='choiceType']",
            "/xs:schema/xs:complexType[@name='groupType']",
            "/xs:schema/xs:complexType[@name='linkType']",
            "/xs:schema/xs:complexType[@name='symbolsType']",
            "/xs:schema/xs:complexType[@name='basicComponent']",
            "/xs:schema/xs:simpleType[@name='validItemName']",
            "/xs:schema/xs:simpleType[@name='validNXClassName']",
            "/xs:schema/xs:simpleType[@name='validTargetName']",
            "/xs:schema/xs:simpleType[@name='nonNegativeUnbounded']",
        )
        for path in path_list:
            nodes = self.pick_nodes_from_xpath(tree, path)
            self._print("\n.. Xpath = %s\n" % path)
            self.general_handler(parent=nodes[0])

        self._print(DATATYPE_POSTAMBLE)

    def _tag_match(self, parent, match_list):
        """match this tag to a list"""
        if parent is None:
            raise ValueError("Must supply a valid parent node")
        parent_tag = parent.tag
        tag_found = False
        for item in match_list:
            # this routine only handles certain XML Schema components
            tag_found = parent_tag == "{%s}%s" % (self.ns["xs"], item)
            if tag_found:
                break
        return tag_found

    def _indent(self, indentLevel):
        return self.INDENTATION * indentLevel

    def print_title(self, title, indentLevel):
        self._print(title)
        self._print(self.TITLE_MARKERS[indentLevel] * len(title) + "\n")

    def general_handler(self, parent=None, indentLevel=0):
        """Handle XML nodes like the former XSLT template"""
        # ignore things we don't know how to handle
        known_tags = ("complexType", "simpleType", "group", "element", "attribute")
        if not self._tag_match(parent, known_tags):
            return

        parent_name = parent.get("name")
        if parent_name is None:
            return

        simple_tag = parent.tag[
            parent.tag.find("}") + 1 :
        ]  # cut off the namespace identifier

        # <varlistentry> ...
        name = parent_name  # + ' data type'
        if simple_tag == "attribute":
            name = "@" + name

        if indentLevel == 0 and simple_tag not in ("attribute"):
            self._print(f".. index:: ! {name} (NXDL data type)\n")
            self._print(f"\n.. _NXDL.data.type.{name}:\n")

        self.print_title(name, indentLevel)

        self.print_docs(parent, indentLevel)

        if len(parent.xpath("xs:attribute", namespaces=self.ns)) > 0:
            self.print_title("Attributes of " + name, indentLevel + 1)
            self.apply_templates(parent, "xs:attribute", indentLevel + 1)

        node_list = parent.xpath("xs:restriction", namespaces=self.ns)
        if len(node_list) > 0:
            # print_title("Restrictions of "+name, indentLevel+1)
            self.restriction_handler(node_list[0], indentLevel + 1)
        node_list = parent.xpath(
            "xs:simpleType/xs:restriction/xs:enumeration", namespaces=self.ns
        )
        if len(node_list) > 0:
            #        print_title("Enumerations of "+name, indentLevel+1)
            self.apply_templates(
                parent,
                "xs:simpleType/xs:restriction",
                indentLevel + 1,
                handler=self.restriction_handler,
            )

        if len(parent.xpath("xs:sequence/xs:element", namespaces=self.ns)) > 0:
            self.print_title("Elements of " + name, indentLevel + 1)
            self.apply_templates(parent, "xs:sequence/xs:element", indentLevel + 1)

        node_list = parent.xpath("xs:sequence/xs:group", namespaces=self.ns)
        if len(node_list) > 0:
            self.print_title("Groups under " + name, indentLevel + 1)
            self.print_docs(node_list[0], indentLevel + 1)

        self.apply_templates(parent, "xs:simpleType", indentLevel + 1)
        self.apply_templates(parent, "xs:complexType", indentLevel + 1)
        self.apply_templates(parent, "xs:complexType/xs:attribute", indentLevel + 1)
        self.apply_templates(
            parent, "xs:complexContent/xs:extension/xs:attribute", indentLevel + 1
        )
        self.apply_templates(
            parent, "xs:complexType/xs:sequence/xs:attribute", indentLevel + 1
        )
        self.apply_templates(
            parent, "xs:complexType/xs:sequence/xs:element", indentLevel + 1
        )
        self.apply_templates(
            parent,
            "xs:complexContent/xs:extension/xs:sequence/xs:element",
            indentLevel + 1,
        )

    def restriction_handler(self, parent=None, indentLevel=0):
        """Handle XSD restriction nodes like the former XSLT template"""
        if not self._tag_match(parent, ("restriction",)):
            return
        self.print_docs(parent, indentLevel)
        self._print("\n")
        self._print(self._indent(indentLevel) + "The value may be any")
        base = parent.get("base")
        pattern_nodes = parent.xpath("xs:pattern", namespaces=self.ns)
        enumeration_nodes = parent.xpath("xs:enumeration", namespaces=self.ns)
        if len(pattern_nodes) > 0:
            self._print(
                self._indent(indentLevel)
                + "``%s``" % base
                + " that *also* matches the regular expression::\n"
            )
            self._print(
                self._indent(indentLevel) + " " * 4 + pattern_nodes[0].get("value")
            )
        elif len(pattern_nodes) > 0:
            # how will this be reached?  Perhaps a deprecated procedure
            self._print(
                self._indent(indentLevel) + "``%s``" % base + " from this list:"
            )
            for node in enumeration_nodes:
                self.enumeration_handler(node, indentLevel)
                self.print_docs(node, indentLevel)
            self._print(self._indent(indentLevel))
        elif len(enumeration_nodes) > 0:
            self._print(self._indent(indentLevel) + "one from this list only:\n")
            for node in enumeration_nodes:
                self.enumeration_handler(node, indentLevel)
                self.print_docs(parent, indentLevel)
            self._print(self._indent(indentLevel))
        else:
            self._print("@" + base)
        self._print("\n")

    def enumeration_handler(self, parent=None, indentLevel=0):
        """Handle XSD enumeration nodes like the former XSLT template"""
        if not self._tag_match(parent, ["enumeration"]):
            return
        self._print(self._indent(indentLevel) + "* ``%s``" % parent.get("value"))
        self.print_docs(parent, indentLevel)

    def apply_templates(self, parent, path, indentLevel, handler=None):
        """iterate the nodes found on the supplied XPath expression"""
        if handler is None:
            handler = self.general_handler
        db = {}
        for node in parent.xpath(path, namespaces=self.ns):
            name = node.get("name") or node.get("ref") or node.get("value")
            if name is not None:
                if name in ("nx:groupGroup",):
                    self._print(">" * 45, name)
                if name in db:
                    raise KeyError("Duplicate name found: " + name)
                db[name] = node
        for name in sorted(db):
            node = db[name]
            handler(node, indentLevel)
            # self.print_docs(node, indentLevel)

    def print_docs(self, parent, indentLevel=0):
        docs = self.get_doc_from_node(parent)
        if docs is not None:
            self._print(self._indent(indentLevel) + "\n")
            for line in docs.splitlines():
                self._print(self._indent(indentLevel) + line)
            self._print(self._indent(indentLevel) + "\n")

    def get_doc_from_node(self, node, retval=None):
        annotation_node = node.find("xs:annotation", self.ns)
        if annotation_node is None:
            return retval
        documentation_node = annotation_node.find("xs:documentation", self.ns)
        if documentation_node is None:
            return retval

        # Be sure to grab _all_ content in the <xs:documentation> node.
        # In the documentation nodes, use XML entities ("&lt;"" instead of "<")
        # for documentation characters that would otherwise be considered as XML.
        s = lxml.etree.tostring(documentation_node, method="text", pretty_print=True)
        rst = s.decode().lstrip("\n")  # remove any leading blank lines
        rst = rst.rstrip()  # remove any trailing white space
        text = textwrap.dedent(rst)  # remove common leading space

        # substitute HTML entities in markup: "<" for "&lt;"
        # thanks: http://stackoverflow.com/questions/2087370/decode-html-entities-in-python-string
        try:  # see #661
            import html

            text = html.unescape(text)
        except (ImportError, AttributeError):
            from html import parser as HTMLParser

            htmlparser = HTMLParser.HTMLParser()
            text = htmlparser.unescape(text)

        return text.lstrip()

    def add_figure(self, name, indentLevel=0):
        imageFile = f"img/nxdl/nxdl_{name}.png"
        figure_id = f"fig.nxdl_{name}"
        file_name = directories.manual_source_sphinxroot() / imageFile
        if not file_name.exists():
            return
        text = FIGURE_FMT % (
            figure_id,
            imageFile,
            name,
            "80%",
            name,
            name,
        )
        indent = self._indent(indentLevel)
        for line in text.splitlines():
            self._print(indent + line)
        self._print("\n")

    def pick_nodes_from_xpath(self, parent, path):
        return parent.xpath(path, namespaces=self.ns)

    def _print(self, *args, end="\n"):
        # TODO: change instances of \t to proper indentation
        self._rst_lines.append(" ".join(args) + end)


ELEMENT_DICT = {
    "attribute": """
An ``attribute`` element can *only* be a child of a
``field`` or ``group`` element.
It is used to define *attribute* elements to be used and their data types
and possibly an enumeration of allowed values.

For more details, see:
:ref:`NXDL.data.type.attributeType`
                """,
    "definition": """
A ``definition`` element can *only* be used
at the root level of an NXDL specification.
Note:  Due to the large number of attributes of the ``definition`` element,
they have been omitted from the figure below.

For more details, see:
:ref:`NXDL.data.type.definition`,
:ref:`NXDL.data.type.definitionType`, and
:ref:`NXDL.data.type.definitionTypeAttr`
                """,
    "dimensions": """
The ``dimensions`` element describes the *shape* of an array.
It is used *only* as a child of a ``field`` element.

For more details, see:
:ref:`NXDL.data.type.dimensionsType`
                """,
    "doc": """
A ``doc`` element can be a child of most NXDL elements.  In most cases, the
content of the ``doc`` element will also become part of the NeXus manual.

:element: {any}:

In documentation, it may be useful to
use an element that is not directly specified by the NXDL language.
The *any* element here says that one can use any element
at all in a ``doc`` element and NXDL will not process it but pass it through.

For more details, see:
:ref:`NXDL.data.type.docType`
                """,
    "enumeration": """
An ``enumeration`` element can *only* be a child of a
``field`` or ``attribute`` element.
It is used to restrict the available choices to a predefined list,
such as to control varieties in spelling of a controversial word (such as
*metre* vs. *meter*).

For more details, see:
:ref:`NXDL.data.type.enumerationType`
                """,
    "field": """
The ``field`` element provides the value of a named item.  Many different attributes
are available to further define the ``field``.  Some of the attributes are not
allowed to be used together (such as ``axes`` and ``axis``); see the documentation
of each for details.
It is used *only* as a child of a ``group`` element.

For more details, see:
:ref:`NXDL.data.type.fieldType`
                """,
    "choice": """
A ``choice`` element is used when a named group might take one
of several possible NeXus base classes.  Logically, it must
have at least two group children.

For more details, see:
:ref:`NXDL.data.type.choiceType`
                """,
    "group": """
A ``group`` element can *only* be a child of a
``definition`` or ``group`` element.
It describes a common level of organization in a NeXus data file, similar
to a subdirectory in a file directory tree.

For more details, see:
:ref:`NXDL.data.type.groupType`
                """,
    "link": """
.. index::
    single: link target

A ``link`` element can *only* be a child of a
``definition``,
``field``, or ``group`` element.
It describes the path to the original source of the parent
``definition``,
``field``, or ``group``.

For more details, see:
:ref:`NXDL.data.type.linkType`
                """,
    "symbols": """
A ``symbols`` element can *only* be a child of a ``definition`` element.
It defines the array index symbols to be used when defining arrays as
``field`` elements with common dimensions and lengths.

For more details, see:
:ref:`NXDL.data.type.symbolsType`
                """,
}


ELEMENT_PREAMBLE = """
=============================
NXDL Elements and Field Types
=============================

The documentation in this section has been obtained directly
from the NXDL Schema file:  *nxdl.xsd*.
First, the basic elements are defined in alphabetical order.
Attributes to an element are indicated immediately following the element
and are preceded with an "@" symbol, such as
**@attribute**.
Then, the common data types used within the NXDL specification are defined.
Pay particular attention to the rules for *validItemName*
and  *validNXClassName*.

..
    2010-11-29,PRJ:
    This contains a lot of special case code to lay out the NXDL chapter.
    It could be cleaner but that would also involve some cooperation on
    anyone who edits nxdl.xsd which is sure to break.  The special case ensures
    the parts come out in the chosen order.  BUT, it is possible that new
    items in nxdl.xsd will not automatically go in the manual.
    Can this be streamlined with some common methods?
    Also, there is probably too much documentation in nxdl.xsd.  Obscures the function.

.. index::
    see:attribute; NXDL attribute
    ! single: NXDL elements

.. _NXDL.elements:

NXDL Elements
=============

    """

DATATYPE_PREAMBLE = """

.. _NXDL.data.types.internal:

NXDL Field Types (internal)
===========================

Field types that define the NXDL language are described here.
These data types are defined in the XSD Schema (``nxdl.xsd``)
and are used in various parts of the Schema to define common structures
or to simplify a complicated entry.  While the data types are not intended for
use in NXDL specifications, they define structures that may be used in NXDL specifications.

"""

DATATYPE_POSTAMBLE = """
**The** ``xs:string`` **data type**
    The ``xs:string`` data type can contain characters,
    line feeds, carriage returns, and tab characters.
    See https://www.w3schools.com/xml/schema_dtypes_string.asp
    for more details.

**The** ``xs:token`` **data type**
    The ``xs:string`` data type is derived from the
    ``xs:string`` data type.

    The ``xs:token`` data type also contains characters,
    but the XML processor will remove line feeds, carriage returns, tabs,
    leading and trailing spaces, and multiple spaces.
    See https://www.w3schools.com/xml/schema_dtypes_string.asp
    for more details.
"""


FIGURE_FMT = """
.. compound::

    .. _%s:

    .. figure:: %s
        :alt: fig.nxdl/nxdl_%s
        :width: %s

        Graphical representation of the NXDL ``%s`` element

    .. Images of NXDL structure are generated from nxdl.xsd source
        using the Eclipse XML Schema Editor (Web Tools Platform).  Open the nxdl.xsd file and choose the
        "Design" tab.  Identify the structure to be documented and double-click to expand
        as needed to show the detail.  Use the XSD > "Export Diagram as Image ..." menu item (also available
        as button in top toolbar).
        Set the name: "nxdl_%s.png" and move the file into the correct location using
        your operating system's commands.  Commit the revision to version control.
"""



#!/usr/bin/env python
"""
Writes a NeXus HDF5 file using h5py with links to data in other HDF5 files.

This example is based on ``writer_2_1``.
"""

from pathlib import Path

import h5py
import numpy

from nexusformat.nexus import (NXdata, NXdetector, NXentry, NXfield,
                               NXinstrument, NXlink, nxopen)

FILE_HDF5_MASTER = "external_master.hdf5"
FILE_HDF5_ANGLES = "external_angles.hdf5"
FILE_HDF5_COUNTS = "external_counts.hdf5"

# ---------------------------

# get some data
filename = str(Path(__file__).absolute().parent.parent / "simple_example.dat")
buffer = numpy.loadtxt(filename).T
tthData = buffer[0]  # float[]
countsData = numpy.asarray(buffer[1], "int32")  # int[]

# put the angle data in an external (non-NeXus) HDF5 data file
with h5py.File(FILE_HDF5_ANGLES, "w") as f:
    ds = f.create_dataset("angles", data=tthData)
    ds.attrs["units"] = "degrees"

# put the detector counts in an external HDF5 data file
# with *incomplete* NeXus structure (no NXdata group)
with nxopen(FILE_HDF5_COUNTS, "w") as f:
    f["entry"] = NXentry()
    f["entry/instrument"] = NXinstrument()
    f["entry/instrument/detector"] = NXdetector()
    f["entry/instrument/detector/counts"] = NXfield(countsData, units="counts")
    f["entry/instrument/detector/two_theta"] = NXlink("/angles",
                                                      FILE_HDF5_ANGLES)

# create a master NeXus HDF5 file
with nxopen(FILE_HDF5_MASTER, "w") as f:
    f["entry"] = NXentry()
    counts = NXlink("/entry/instrument/detector/counts", FILE_HDF5_COUNTS,
                    name="counts")
    two_theta = NXlink("/angles", FILE_HDF5_ANGLES, name="two_theta")
    f["entry/data"] = NXdata(counts, two_theta)
    f["entry/data"].set_default()
    f["entry/instrument"] = NXlink("/entry/instrument", FILE_HDF5_COUNTS)



#!/usr/bin/env python

"""
update the copyright date in all NeXus text files

This is the bash command to find all matching lines::

  grep -iR copyright | grep -i "(c)" | grep -i nexus

See copyright text at bottom of this file for example.
"""

import os, sys
import mimetypes
from build_preparation import ROOT_DIR_EXPECTED_RESOURCES
import datetime

YEAR = datetime.datetime.now().year
LEFT_SIDE_TEXT_MATCH = "Copyright (C) "
RIGHT_SIDE_TEXT_MATCH = " NeXus International Advisory Committee (NIAC)"


def update(filename):
    """update the copyright year in the file"""

    def position(line, key):
        pos = line.find(key)
        if pos >= 0:
            if line.find("NIAC") >= 0:
                return pos
        return None

    if not os.path.exists(filename):
        return
    changes = []
    buf = open(filename).readlines()
    for number, line in enumerate(buf):
        pos = position(line, LEFT_SIDE_TEXT_MATCH)
        if pos is None:
            continue  # no match

        pos += len(LEFT_SIDE_TEXT_MATCH)
        text_l = line[:pos]

        text = line[pos:]
        pos = text.find(RIGHT_SIDE_TEXT_MATCH)
        text_r = text[pos:]

        try:
            years = list(map(int, text[:pos].split("-")))
            if len(years) in (1, 2):
                if len(years) == 1:
                    years.append(YEAR)
                elif len(years) == 2:
                    years[1] = YEAR
                line_new = text_l + "-".join(map(str, years)) + text_r
                changes.append(list((number, line_new)))
        except Exception as _exc:
            print(number, filename, str(_exc))
    for number, line in changes:
        buf[number] = line
    if len(changes) > 0:
        print("Update: ", filename)
        fp = open(filename, "w")
        fp.writelines(buf)
        fp.close()


def find_source_files(path):
    """walk the source_path directories accumulating files to be checked"""
    file_list = []
    for root, dirs, files in os.walk(path):
        if root.find("/.git") < 0 or root.find("/kits") < 0:
            file_list = file_list + [os.path.join(root, _) for _ in files]
    return file_list


def sift_file_list(file_list):
    """remove known non-text files and paths"""
    new_list = []
    acceptable_mime_types = """
        application/xml
        application/x-msdos-program
        application/xslt+xml
    """.strip().split()
    ignore_extensions = """
    .dia .vsdx .h5 .nx .hdf5 .hdf .nx5 .pyc
    """.strip().split()
    for fn in file_list:
        _fn = os.path.split(fn)[-1]
        mime = mimetypes.guess_type(fn)[0]
        if fn.find("/.git") >= 0:
            continue
        if fn.find("/.settings") >= 0:
            continue
        if fn.find("/kits") >= 0:
            continue
        if fn.find("/build") >= 0:
            continue
        if os.path.splitext(fn)[-1] in ignore_extensions:
            continue
        if mime is None or mime.startswith("text/") or mime in acceptable_mime_types:
            new_list.append(fn)
    return new_list


def is_definitions_directory(basedir):
    """test if ``basedir`` is a NeXus definitions directory"""
    # look for the expected files and subdirectories in the root directory
    for item_list in ROOT_DIR_EXPECTED_RESOURCES.values():
        for item in item_list:
            if not os.path.exists(os.path.join(basedir, item)):
                return False
    return True


def qualify_inputs(root_dir):
    """raise error if this program cannot continue, based on the inputs"""
    if not os.path.exists(root_dir):
        raise RuntimeError("Cannot find " + root_dir)

    if not os.path.isdir(root_dir):
        raise RuntimeError("Not a directory: " + root_dir)

    if not is_definitions_directory(root_dir):
        msg = "Not a NeXus definitions root directory " + root_dir
        raise RuntimeError(msg)


def command_args():
    """get the command-line arguments, handle syntax errors"""
    import argparse

    doc = __doc__.strip().splitlines()[0]
    parser = argparse.ArgumentParser(prog=sys.argv[0], description=doc)
    parser.add_argument(
        "defs_dir", action="store", help="NeXus definitions root directory"
    )
    return parser.parse_args()


def main():
    """
    standard command-line processing
    
    source directory (NeXus definitions dir) named as command line argument
    target directory is specified (or defaults to present working directory)
    """
    cli = command_args()
    root_dir = os.path.abspath(cli.defs_dir)
    qualify_inputs(root_dir)

    file_list = sift_file_list(find_source_files(root_dir))
    for fn in file_list:
        update(fn)


def __developer_build_setup__():
    """for use with source-code debugger ONLY"""
    import shutil

    # sys.argv.append('-h')
    sys.argv.append("..")


if __name__ == "__main__":
    # __developer_build_setup__()
    main()


# NeXus - Neutron and X-ray Common Data Format
#
# Copyright (C) 2008-2022 NeXus International Advisory Committee (NIAC)
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 3 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
# For further information, see http://www.nexusformat.org



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Classes representing groups with NeXus-ish formatted data parsed from hspy."""

# pylint: disable=no-member

from typing import Dict

import numpy as np

import hyperspy.api as hs

from pynxtools.dataconverter.readers.em_spctrscpy.utils.em_nexus_base_classes \
    import NxObject


class HspyRectRoiEelsAllSpectra:
    """Representing a regular stack of EELS spectra over a rectangular ROI."""

    def __init__(self, hspy_clss):
        self.meta: Dict[str, NxObject] = {}
        self.meta["program"] = NxObject(value="hyperspy")
        self.meta["program_version"] = NxObject(value=hs.__version__)
        self.meta["title"] = NxObject()
        self.meta["long_name"] = NxObject()  # value="counts")
        self.meta["counts"] = NxObject()
        self.meta["xpos"] = NxObject()
        self.meta["xpos_long_name"] = NxObject()
        self.meta["ypos"] = NxObject()
        self.meta["ypos_long_name"] = NxObject()
        self.meta["energy_loss"] = NxObject()
        self.meta["energy_loss_long_name"] = NxObject()
        # value="Electron energy loss")
        self.is_valid = True

        self.is_supported(hspy_clss)
        self.parse(hspy_clss)

    def is_supported(self, hspy_s3d):
        """Check if the input has supported axes_manager and key metadata."""
        assert hspy_s3d.metadata["Signal"]["signal_type"] == "EELS", \
            "hspy_s3d is not a valid EELS hyperspy instance !"
        assert hspy_s3d.data.ndim == 3, \
            "hspy_s3d is not a valid 3D dataset !"
        axes_dict = hspy_s3d.axes_manager.as_dictionary()
        required_axis_names = ["axis-0", "axis-1", "axis-2"]
        for req_key in required_axis_names:
            assert req_key in axes_dict.keys(), \
                req_key + " is unexpectedly not registered in the axes_manager !"
        required_keywords = ["_type", "name", "units", "size", "scale", "offset"]
        avail_axis_names = []
        for keyword in axes_dict.keys():
            for req_key in required_keywords:  # check if all required keys exist
                assert req_key in axes_dict[keyword].keys(), \
                    "hspy_s3d axis " + keyword + " lacks " + req_key + " !"

            assert axes_dict[keyword]["_type"] == "UniformDataAxis", \
                keyword + ", this axis is not of type UniformDataAxis !"
            avail_axis_names.append(axes_dict[keyword]["name"])

        print(np.sort(avail_axis_names))
        print(np.sort(["y", "x", "Energy loss"]))
        axes_as_expected = np.all(
            np.sort(avail_axis_names) == np.sort(["y", "x", "Energy loss"]))
        if axes_as_expected is False:
            print(f"\tIn function {__name__} as expected")
            self.is_valid = False

    def parse(self, hspy_s3d):
        """Parse a hyperspy Signal3D stack instance into an NX default plottable."""
        if self.is_valid is False:
            pass
        print(f"\tIn function {__name__}")
        # self.meta["title"].value = hspy_s3d.metadata["General"]["title"]
        self.meta["long_name"].value = hspy_s3d.metadata["Signal"]["signal_type"]
        # self.meta["long_name"].value = hspy_s3d.metadata["General"]["title"]
        self.meta["title"].value = hspy_s3d.metadata["Signal"]["signal_type"]
        self.meta["counts"].value = hspy_s3d.data  # hspy uses numpy and adapts ??
        axes_dict = hspy_s3d.axes_manager.as_dictionary()
        for keyword in axes_dict.keys():
            offset = np.float64(axes_dict[keyword]["offset"])
            scale = np.float64(axes_dict[keyword]["scale"])
            size = np.uint32(axes_dict[keyword]["size"])
            unit = str(axes_dict[keyword]["units"])
            if axes_dict[keyword]["name"] == "y":
                self.meta["ypos"].value = np.asarray(
                    np.linspace(0., np.float64(size) * scale, num=size,
                                endpoint=True) + offset / 2., np.float64)
                self.meta["ypos"].unit = unit
                self.meta["ypos_long_name"].value = "y"
            elif axes_dict[keyword]["name"] == "x":
                self.meta["xpos"].value = np.asarray(
                    np.linspace(0., np.float64(size) * scale, num=size,
                                endpoint=True) + offset / 2., np.float64)
                self.meta["xpos"].unit = unit
                self.meta["xpos_long_name"].value = "x"
            else:  # axes_dict[keyword]["name"] == "Energy loss":
                self.meta["energy_loss"].value = np.asarray(
                    np.linspace(0., np.float64(size) * scale, num=size,
                                endpoint=True) + offset / 2., np.float64)
                self.meta["energy_loss"].unit = unit
                self.meta["energy_loss_long_name"].value = "Energy loss"


class HspyRectRoiEelsSummarySpectrum:
    """Represent/compute an accumulated EELS spectrum from hspy EELSSpectrum."""

    def __init__(self, hspy_clss):
        self.meta: Dict[str, NxObject] = {}
        self.meta["program"] = NxObject(value="hyperspy")
        self.meta["program_version"] = NxObject(value=hs.__version__)
        self.meta["title"] = NxObject()
        self.meta["long_name"] = NxObject()  # value="Counts")
        self.meta["counts"] = NxObject()
        self.meta["energy_loss"] = NxObject()
        self.meta["energy_loss_long_name"] = NxObject()  # value="Energy loss")
        self.is_valid = True

        self.is_supported(hspy_clss)
        self.parse(hspy_clss)

    def is_supported(self, hspy_s3d):
        """Check if the input has supported axes_manager and key metadata."""
        assert hspy_s3d.metadata["Signal"]["signal_type"] == "EELS", \
            "hspy_s3d is not a valid EELS hyperspy instance !"
        assert hspy_s3d.data.ndim == 3, \
            "hspy_s3d is not a valid 3D dataset !"
        axes_dict = hspy_s3d.axes_manager.as_dictionary()
        required_axis_names = ["axis-0", "axis-1", "axis-2"]
        for req_key in required_axis_names:
            assert req_key in axes_dict.keys(), \
                req_key + " is unexpectedly not registered in the axes_manager !"
        required_keywords = ["_type", "name", "units", "size", "scale", "offset"]
        avail_axis_names = []
        for keyword in axes_dict.keys():
            for req_key in required_keywords:  # check if all required keys exist
                assert req_key in axes_dict[keyword].keys(), \
                    "hspy_s3d axis " + keyword + " lacks " + req_key + " !"

            assert axes_dict[keyword]["_type"] == "UniformDataAxis", \
                keyword + ", this axis is not of type UniformDataAxis !"
            avail_axis_names.append(axes_dict[keyword]["name"])

        print(np.sort(avail_axis_names))
        print(np.sort(["y", "x", "Energy loss"]))
        axes_as_expected = np.all(
            np.sort(avail_axis_names) == np.sort(["y", "x", "Energy loss"]))
        if axes_as_expected is False:
            print(f"\tIn function {__name__} as expected")
            self.is_valid = False

    def parse(self, hspy_s3d):
        """Summarize the spectra stack into a ROI summary spectrum."""
        if self.is_valid is False:
            pass
        print(f"\tIn function {__name__}")
        self.meta["long_name"].value = hspy_s3d.metadata["Signal"]["signal_type"]
        self.meta["title"].value = hspy_s3d.metadata["Signal"]["signal_type"]
        axes_dict = hspy_s3d.axes_manager.as_dictionary()
        # ##MK::assume for now that hspy arranges the axes such that the
        # Energy loss comes last, then we can iterate over
        # ##MK::at least in the example of DM3 files I found that the
        # EELSSpectrum.data array is float32 this should not be done
        # inspecting np.unique(hspy_s1d.data[y,x, :]) reveals there is a
        # cast from what was likely an uint(16?) to float32
        # ##MK::for now cast them back to uint32 and accumulate
        # instead it should be uint, so we have to be careful with not
        shape = np.shape(hspy_s3d.data)
        self.meta["counts"].value = np.zeros([shape[2]], np.uint32)
        for y_pixel in np.arange(0, shape[0]):
            for x_pixel in np.arange(1, shape[1]):
                self.meta["counts"].value \
                    += np.asarray(hspy_s3d.data[y_pixel, x_pixel, :], np.uint32)
        # seems that hspy is adaptive, uses numpy under the hood
        # though, so a hspy signal's .data member is already a proper numpy dtype
        # therefore, an explicit call like this
        # np.asarray(hspy_s1d.data, np.uint32) is not necessary
        # ##MK::on the contrary the above-mentioned observed cast to float32
        # means it might not be adaptive enough ??
        for keyword in axes_dict.keys():
            offset = np.float64(axes_dict[keyword]["offset"])
            scale = np.float64(axes_dict[keyword]["scale"])
            size = np.uint32(axes_dict[keyword]["size"])
            unit = str(axes_dict[keyword]["units"])
            if axes_dict[keyword]["name"] == "Energy loss":
                self.meta["energy_loss"].value = np.asarray(
                    np.linspace(0., np.float64(size) * scale, num=size,
                                endpoint=True) + offset / 2., np.float64)
                self.meta["energy_loss"].unit = unit
                self.meta["energy_loss_long_name"].value = "Energy loss"


class NxSpectrumSetEmEels:
    """Representing a set of EELS spectra with metadata."""

    def __init__(self, hspy_list):
        self.stack_data = []  # the HspyRectRoiEelsAllSpectra
        self.summary_data = []  # the HspyRectRoiEelsSummarySpectrum
        self.is_valid = True

        self.is_an_implemented_case(hspy_list)
        self.parse_hspy_instances(hspy_list)

    def is_an_implemented_case(self, hspy_list):
        """Check if signal instances in a list is a supported combination."""
        # an EELS measurement which this implementation supports consist of
        # a rectangular ROI are currently supported, this ROI should have
        # one SpectrumStack cnts = f(y, x, energy loss)
        cardinality_stack = 0
        for hspy_clss in hspy_list:
            if isinstance(hspy_clss, hs.signals.EELSSpectrum) is True:
                assert hspy_clss.data.ndim in [3], \
                    "Unexpectedly found unsupported-dimensional EELSSpectrum!"
                if hspy_clss.data.ndim == 3:
                    cardinality_stack += 1
        if cardinality_stack != 1:
            self.is_valid = False

    def parse_hspy_instances(self, hspy_list):
        """Extract from hspy class instances what NOMAD OASIS understands."""
        if self.is_valid is False:
            pass
        print(f"\tIn function {__name__}")
        for hspy_clss in hspy_list:
            if isinstance(hspy_clss, hs.signals.EELSSpectrum) is True:
                assert hspy_clss.data.ndim in [3], \
                    "Unexpectedly found unsupported-dimensional EELSSpectrum!"
                if hspy_clss.data.ndim == 3:
                    self.stack_data.append(
                        HspyRectRoiEelsAllSpectra(hspy_clss))
                    self.summary_data.append(
                        HspyRectRoiEelsSummarySpectrum(hspy_clss))

    def report(self, prefix: str, frame_id: int,
               ifo: dict, template: dict) -> dict:
        """Enter data from the NX-specific representation into the template."""
        if self.is_valid is False:
            print(f"\t{__name__} reporting nothing!")
            return template
        print(f"\t{__name__} reporting...")
        assert (len(self.stack_data) >= 0) and (len(self.stack_data) <= 1), \
            "More than one spectrum stack is currently not supported!"
        assert (len(self.summary_data) >= 0) and (len(self.summary_data) <= 1), \
            "More than one sum spectrum stack is currently not supported!"

        if len(self.stack_data) == 1:
            trg = f"{prefix}eels/PROCESS[process1]/"
            template[f"{trg}PROGRAM[program1]/program"] = "hyperspy"
            template[f"{trg}PROGRAM[program1]/program/@version"] = hs.__version__
            template[f"{trg}mode"] = "n/a"
            template[f"{trg}detector_identifier"] = "n/a"
            template[f"{trg}source"] = ifo["source_file_name"]
            template[f"{trg}source/@version"] = ifo["source_file_version"]

            trg = f"{prefix}eels/stack/"
            template[f"{trg}title"] = "EELS spectra stack"
            # template[f"{trg}@long_name"] = self.stack_data[0].meta["long_name"].value
            template[f"{trg}@signal"] = "data_counts"
            template[f"{trg}@axes"] = ["axis_y", "axis_x", "axis_energy_loss"]
            template[f"{trg}@AXISNAME[axis_energy_loss_indices]"] = np.uint32(2)
            template[f"{trg}@AXISNAME[axis_x_indices]"] = np.uint32(1)
            template[f"{trg}@AXISNAME[axis_y_indices]"] = np.uint32(0)
            template[f"{trg}DATA[data_counts]"] \
                = {"compress": self.stack_data[0].meta["counts"].value,
                   "strength": 1}
            template[f"{trg}DATA[data_counts]/@units"] = ""
            template[f"{trg}DATA[data_counts]/@long_name"] = "Signal (a.u.)"
            template[f"{trg}AXISNAME[axis_energy_loss]"] \
                = {"compress": self.stack_data[0].meta["energy_loss"].value,
                   "strength": 1}
            template[f"{trg}AXISNAME[axis_energy_loss]/@units"] \
                = self.stack_data[0].meta["energy_loss"].unit
            template[f"{trg}AXISNAME[axis_energy_loss]/@long_name"] \
                = f"Electron energy loss ({self.stack_data[0].meta['energy_loss'].unit})"
            template[f"{trg}AXISNAME[axis_x]"] \
                = {"compress": self.stack_data[0].meta["xpos"].value,
                   "strength": 1}
            template[f"{trg}AXISNAME[axis_x]/@units"] \
                = self.stack_data[0].meta["xpos"].unit
            template[f"{trg}AXISNAME[axis_x]/@long_name"] \
                = f"x ({self.stack_data[0].meta['xpos'].unit})"
            template[f"{trg}AXISNAME[axis_y]"] \
                = {"compress": self.stack_data[0].meta["ypos"].value,
                   "strength": 1}
            template[f"{trg}AXISNAME[axis_y]/@units"] \
                = self.stack_data[0].meta["ypos"].unit
            template[f"{trg}AXISNAME[axis_y]/@long_name"] \
                = f"y ({self.stack_data[0].meta['ypos'].unit})"

        if len(self.summary_data) == 1:
            trg = f"{prefix}eels/summary/"
            template[f"{trg}title"] = "Accumulated EELS spectrum"
            # template[f"{trg}@long_name"] = self.summary_data[0].meta["long_name"].value
            template[f"{trg}@signal"] = "data_counts"
            template[f"{trg}@axes"] = ["axis_energy_loss"]
            template[f"{trg}@AXISNAME[axis_energy_loss_indices]"] = np.uint32(0)
            template[f"{trg}DATA[data_counts]"] \
                = {"compress": self.summary_data[0].meta["counts"].value,
                   "strength": 1}
            template[f"{trg}DATA[data_counts]/@long_name"] = "Signal (a.u.)"
            template[f"{trg}AXISNAME[axis_energy_loss]"] \
                = {"compress": self.summary_data[0].meta["energy_loss"].value,
                   "strength": 1}
            template[f"{trg}AXISNAME[axis_energy_loss]/@units"] \
                = self.summary_data[0].meta["energy_loss"].unit
            template[f"{trg}AXISNAME[axis_energy_loss]/@long_name"] \
                = f"Energy loss ({self.summary_data[0].meta['energy_loss'].unit})"

        return template



#!/usr/bin/env python
"""
Writes the simplest NeXus HDF5 file using h5py

Uses method accepted at 2014NIAC
according to the example from Figure 1.3
in the Introduction chapter
"""

from pathlib import Path

import numpy

from nexusformat.nexus import NXdata, NXentry, NXfield, nxopen

filename = str(Path(__file__).absolute().parent.parent / "simple_example.dat")
buffer = numpy.loadtxt(filename).T
tthData = buffer[0]
countsData = numpy.asarray(buffer[1], "int32")

with nxopen("simple_example_write1.hdf5", "w") as f:  # create the NeXus file
    f["Scan"] = NXentry()
    tth = NXfield(tthData, name="two_theta", units="degrees")
    counts = NXfield(countsData, name="counts", units="counts")
    f["Scan/data"] = NXdata(counts, tth)



from ..globals import directories


def dir_args(parser):
    parser.add_argument(
        "--source-root",
        type=str,
        default=None,
        help="Source root directory" f" Default: {directories.get_source_root()}",
    )
    parser.add_argument(
        "--nxdl-root",
        type=str,
        default=None,
        help="NXDL root directory" f" Default: {directories.get_nxdl_root()}",
    )
    parser.add_argument(
        "--build-root",
        type=str,
        default=None,
        help="Build root directory" f" Default: {directories.get_build_root()}",
    )


def dir_exec(args):
    if args.source_root:
        directories.set_source_root(args.source_root)
    if args.nxdl_root:
        directories.set_nxdl_root(args.nxdl_root)
    if args.build_root:
        directories.set_build_root(args.build_root)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Generator for NXapm default plots."""

# pylint: disable=no-member

import numpy as np

from pynxtools.dataconverter.readers.shared.shared_utils \
    import get_repo_last_commit


def create_default_plot_reconstruction(template: dict, entry_id: int) -> dict:
    """Compute on-the-fly, add, and give path to discretized reconstruction."""
    trg = f"/ENTRY[entry{entry_id}]/atom_probe/reconstruction/"
    xyz = template[f"{trg}reconstructed_positions"]["compress"]

    print(f"\tEnter histogram computation, np.shape(xyz) {np.shape(xyz)}")
    resolution = 1.0  # cubic voxel edge length in nm
    bounds = np.zeros([3, 2], np.float32)  # in nm
    for i in np.arange(0, 3):
        bounds[i, 0] = np.min(xyz[:, i])
        bounds[i, 1] = np.max(xyz[:, i])
    # make the bounding box a quadric prism
    imi = np.floor(bounds[0, 0]) - resolution
    imx = np.ceil(bounds[0, 1]) + resolution
    xedges = np.linspace(imi, imx, num=int(np.ceil((imx - imi) / resolution)) + 1,
                         endpoint=True)
    # this partitioning is not general enough, imi and imx should be left and right
    # bounds respectively
    imi = np.floor(bounds[1, 0]) - resolution
    imx = np.ceil(bounds[1, 1]) + resolution
    yedges = np.linspace(imi, imx, num=int(np.ceil((imx - imi) / resolution)) + 1,
                         endpoint=True)
    imi = np.floor(bounds[2, 0]) - resolution
    imx = np.ceil(bounds[2, 1]) + resolution
    zedges = np.linspace(imi, imx,
                         num=int(np.ceil((imx - imi) / resolution)) + 1,
                         endpoint=True)

    hist3d = np.histogramdd((xyz[:, 0], xyz[:, 1], xyz[:, 2]),
                            bins=(xedges, yedges, zedges))
    del xyz
    assert isinstance(hist3d[0], np.ndarray), \
        "Hist3d computation from the reconstruction failed!"
    assert len(np.shape(hist3d[0])) == 3, \
        "Hist3d computation from the reconstruction failed!"
    for i in np.arange(0, 3):
        assert np.shape(hist3d[0])[i] > 0, \
            "Dimensions " + str(i) + " has no length!"

    trg = f"/ENTRY[entry{entry_id}]/atom_probe/reconstruction/" \
          f"naive_point_cloud_density_map/"
    template[f"{trg}PROGRAM[program1]/program"] = "nomad-parser-nexus/apm/reader.py"
    template[f"{trg}PROGRAM[program1]/program/@version"] = get_repo_last_commit()

    trg = f"/ENTRY[entry{entry_id}]/atom_probe/reconstruction/" \
          f"naive_point_cloud_density_map/DATA[data]/"
    template[f"{trg}title"] = "Discretized reconstruction space"
    # template[f"{trg}@long_name"] = "Discretized reconstruction space"
    template[f"{trg}@signal"] = "data_counts"
    template[f"{trg}@axes"] = ["axis_x", "axis_y", "axis_z"]
    template[f"{trg}@AXISNAME_indices[axis_x]"] = np.uint32(0)
    template[f"{trg}@AXISNAME_indices[axis_y]"] = np.uint32(1)
    template[f"{trg}@AXISNAME_indices[axis_z]"] = np.uint32(2)

    # mind that histogram does not follow Cartesian conventions so a transpose
    # might be necessary, for now we implement the transpose in the appdef
    template[f"{trg}DATA[data_counts]"] \
        = {"compress": np.array(hist3d[0], np.uint32), "strength": 1}
    template[f"{trg}AXISNAME[axis_x]"] \
        = {"compress": np.array(hist3d[1][0][1::], np.float32), "strength": 1}
    template[f"{trg}AXISNAME[axis_x]/@units"] = "nm"
    template[f"{trg}AXISNAME[axis_x]/@long_name"] = "x (nm)"
    template[f"{trg}AXISNAME[axis_y]"] \
        = {"compress": np.array(hist3d[1][1][1::], np.float32), "strength": 1}
    template[f"{trg}AXISNAME[axis_y]/@units"] = "nm"
    template[f"{trg}AXISNAME[axis_y]/@long_name"] = "y (nm)"
    template[f"{trg}AXISNAME[axis_z]"] \
        = {"compress": np.array(hist3d[1][2][1::], np.float32), "strength": 1}
    template[f"{trg}AXISNAME[axis_z]/@units"] = "nm"
    template[f"{trg}AXISNAME[axis_z]/@long_name"] = "z (nm)"
    print("Default plot 3D discretized reconstruction at 1 nm binning.")
    del hist3d

    return template


def create_default_plot_mass_spectrum(template: dict, entry_id: int) -> dict:
    """Compute on-the-fly, add, and give path to discretized reconstruction."""
    trg = f"/ENTRY[entry{entry_id}]/atom_probe/mass_to_charge_conversion/"
    m_z = template[f"{trg}mass_to_charge"]["compress"]

    print(f"\tEnter mass spectrum computation, np.shape(m_z) {np.shape(m_z)}")
    mqmin = 0.0  # in Da, do not plot unphysical values < 0.0
    mqmax = np.ceil(np.max(m_z[:]))
    mqincr = 0.01  # in Da by default

    hist1d = np.histogram(
        m_z[:],
        np.linspace(mqmin, mqmax,
                    num=int(np.ceil((mqmax - mqmin) / mqincr)) + 1,
                    endpoint=True))
    del m_z
    assert isinstance(hist1d[0], np.ndarray), \
        "Hist1d computation from the mass spectrum failed!"
    assert len(np.shape(hist1d[0])) == 1, \
        "Hist1d computation from the mass spectrum failed!"
    for i in np.arange(0, 1):
        assert np.shape(hist1d[0])[i] > 0, \
            "Dimensions " + str(i) + " has no length!"

    trg = f"/ENTRY[entry{entry_id}]/atom_probe/ranging/mass_to_charge_distribution/"
    template[f"{trg}PROGRAM[program1]/program"] = "nomad-parser-nexus/apm/reader.py"
    template[f"{trg}PROGRAM[program1]/program/@version"] = get_repo_last_commit()

    template[f"{trg}range_increment"] = mqincr
    template[f"{trg}range_increment/@units"] = "Da"
    template[f"{trg}range_minmax"] = np.array([mqmin, mqmax], np.float32)
    template[f"{trg}range_minmax/@units"] = "Da"

    trg = f"/ENTRY[entry{entry_id}]/atom_probe/ranging/" \
          f"mass_to_charge_distribution/mass_spectrum/"
    template[f"{trg}title"] = "Mass spectrum (0.01 Da binning)"
    template[f"{trg}@signal"] = "data_counts"
    template[f"{trg}@axes"] = "axis_mass_to_charge"
    template[f"{trg}@AXISNAME_indices[axis_mass_to_charge]"] = np.uint32(0)
    template[f"{trg}DATA[data_counts]"] \
        = {"compress": np.array(hist1d[0], np.uint32), "strength": 1}
    template[f"{trg}DATA[data_counts]/@long_name"] = "Counts (1)"
    template[f"{trg}AXISNAME[axis_mass_to_charge]"] \
        = {"compress": np.array(hist1d[1][1::], np.float32), "strength": 1}
    template[f"{trg}AXISNAME[axis_mass_to_charge]/@units"] = "Da"
    template[f"{trg}AXISNAME[axis_mass_to_charge]/@long_name"] \
        = "Mass-to-charge-state ratio (Da)"
    print("Plot mass spectrum at 0.01 Da binning was created.")
    del hist1d

    return template


def apm_default_plot_generator(template: dict, n_entries: int) -> dict:
    """Copy data from self into template the appdef instance."""
    print("Create default plots on-the-fly...")
    # now the reader implements what is effectively the task of a normalizer step
    # adding plot (discretized representation of the dataset), for now the default plot
    # adding plot mass-to-charge-state ratio histogram,
    # termed mass spectrum in APM community

    # NEW ISSUE: add path to default plottable data

    # check if reconstructed ion positions have been stored
    for entry_id in np.arange(1, n_entries + 1):
        trg = f"/ENTRY[entry{entry_id}]/atom_probe/mass_to_charge_conversion/"
        has_valid_m_z = False
        path = f"{trg}mass_to_charge"
        if isinstance(template[f"{trg}mass_to_charge"], dict):
            if "compress" in template[path].keys():
                if isinstance(template[path]["compress"], np.ndarray):
                    has_valid_m_z = True

        trg = f"/ENTRY[entry{entry_id}]/atom_probe/reconstruction/"
        has_valid_xyz = False
        path = f"{trg}reconstructed_positions"
        if isinstance(template[path], dict):
            if "compress" in template[path].keys():
                if isinstance(template[path]["compress"], np.ndarray):
                    has_valid_xyz = True

        has_default_data = has_valid_m_z or has_valid_xyz
        assert has_default_data is True, \
            "Having no recon or mass-to-charge data is inacceptable at the moment!"

        # NEW ISSUE: fall-back solution to plot something else, however
        # currently POS, EPOS and APT provide always xyz, and m_z data

        # generate default plottable and add path
        template["/@default"] = f"entry{entry_id}"
        trg = f"/ENTRY[entry{entry_id}]/"
        template[f"{trg}@default"] = "atom_probe"

        if has_valid_m_z is True:
            create_default_plot_mass_spectrum(template, entry_id)
            # mass_spectrum main default...
            trg += "atom_probe/"
            template[f"{trg}@default"] = "ranging"
            trg += "ranging/"
            template[f"{trg}@default"] = "mass_to_charge_distribution"
            trg += "mass_to_charge_distribution/"
            template[f"{trg}@default"] = "mass_spectrum"

        if has_valid_xyz is True:
            # ... discretized naive tomographic reconstruction as fallback...
            create_default_plot_reconstruction(template, entry_id)
            # generate path to the default plottable
            if has_valid_m_z is False:
                trg += "atom_probe/"
                template[f"{trg}@default"] = "reconstruction"
                trg += "reconstruction/"
                template[f"{trg}@default"] = "naive_point_cloud_density_map"
                trg += "naive_point_cloud_density_map/"
                template[f"{trg}@default"] = "data"

    return template



"""Namespace of the schema in which NXDL is defined
(the XSD namespace) and the namespace of the NeXus
class definitions (the NXDL namespace).

Namespaces are URL's solely to be globally unique.
Do not use these URL's for validation. That's what
"xsi:schemaLocation" is for. Currently this is the
relative URI "../nxdl.xsd" which means validation
can only be done for NXDL files in subdirectories.
"""

from .directories import get_nxdl_version_file

XSD_NAMESPACE = "http://www.w3.org/2001/XMLSchema"
NXDL_NAMESPACE = "http://definition.nexusformat.org/nxdl/3.1"


def get_nxdl_version() -> str:
    """The version of the NeXus standard and the NeXus Definition language"""
    with open(get_nxdl_version_file(), "r") as fh:
        return fh.read().strip()



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Utilities for working with NeXus concepts encoded as Python dicts in the concepts dir."""

# pylint: disable=no-member

import pytz

from datetime import datetime


def load_from_modifier(terms, fd_dct):
    """Implement modifier which reads values of different type from fd_dct."""
    if isinstance(terms, str):
        if terms in fd_dct.keys():
            return fd_dct[terms]
    if all(isinstance(entry, str) for entry in terms) is True:
        if isinstance(terms, list):
            lst = []
            for entry in terms:
                lst.append(fd_dct[entry])
            return lst
    return None


def convert_iso8601_modifier(terms, dct: dict):
    """Implement modifier which transforms nionswift time stamps to proper UTC ISO8601."""
    if terms is not None:
        if isinstance(terms, str):
            if terms in dct.keys():
                return None
        elif (isinstance(terms, list)) and (len(terms) == 2) \
                and (all(isinstance(entry, str) for entry in terms) is True):
            # assume the first argument is a local time
            # assume the second argument is a timezone string
            if terms[0] in dct.keys() and terms[1] in dct.keys():
                # handle the case that these times can be arbitrarily formatted
                # for now we let ourselves be guided
                # by how time stamps are returned in Christoph Koch's
                # nionswift instances also formatting-wise
                date_time_str = dct[terms[0]].replace("T", " ")
                time_zone_str = dct[terms[1]]
                if time_zone_str in pytz.all_timezones:
                    date_time_obj \
                        = datetime.strptime(date_time_str, '%Y-%m-%d %H:%M:%S.%f')
                    utc_time_zone_aware \
                        = pytz.timezone(time_zone_str).localize(date_time_obj)
                    return utc_time_zone_aware
                else:
                    raise ValueError('Invalid timezone string!')
                return None
        else:
            return None
    return None


def apply_modifier(modifier, dct: dict):
    """Interpret a functional mapping using data from dct via calling modifiers."""
    if isinstance(modifier, dict):
        # different commands are available
        if set(["fun", "terms"]) == set(modifier.keys()):
            if modifier["fun"] == "load_from":
                return load_from_modifier(modifier["terms"], dct)
            if modifier["fun"] == "convert_iso8601":
                return convert_iso8601_modifier(modifier["terms"], dct)
        elif set(["link"]) == set(modifier.keys()):
            # CURRENTLY NOT IMPLEMENTED
            # with the jsonmap reader Sherjeel conceptualized "link"
            return None
        else:
            return None
    if isinstance(modifier, str):
        return modifier
    return None


# examples/tests how to use modifiers
# modd = "µs"
# modd = {"link": "some_link_to_somewhere"}
# modd = {"fun": "load_from", "terms": "metadata/scan/scan_device_properties/mag_boards/MagBoard 1 DAC 11"}
# modd = {"fun": "load_from", "terms": ["metadata/scan/scan_device_properties/mag_boards/MagBoard 1 DAC 11",
#     "metadata/scan/scan_device_properties/mag_boards/MagBoard 1 Relay"]}
# modd = {"fun": "convert_iso8601", "terms": ["data_modified", "timezone"]}
# print(apply_modifier(modd, yml))

def variadic_path_to_specific_path(path: str, instance_identifier: list):
    """Transforms a variadic path to an actual path with instances."""
    if (path is not None) and (path != ""):
        narguments = path.count("*")
        if narguments == 0:  # path is not variadic
            return path
        if len(instance_identifier) >= narguments:
            tmp = path.split("*")
            if len(tmp) == narguments + 1:
                nx_specific_path = ""
                for idx in range(0, narguments):
                    nx_specific_path += f"{tmp[idx]}{instance_identifier[idx]}"
                    idx += 1
                nx_specific_path += f"{tmp[-1]}"
                return nx_specific_path
    return None



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Parser for loading generic orientation microscopy data based on ."""

# pylint: disable=no-member

from typing import Tuple, Any

from pynxtools.dataconverter.readers.base.reader import BaseReader

from pynxtools.dataconverter.readers.em_om.utils.use_case_selector \
    import EmOmUseCaseSelector

from pynxtools.dataconverter.readers.em_om.utils.generic_eln_io \
    import NxEmOmGenericElnSchemaParser

from pynxtools.dataconverter.readers.em_om.utils.orix_ebsd_parser \
    import NxEmOmOrixEbsdParser

from pynxtools.dataconverter.readers.em_om.utils.mtex_ebsd_parser \
    import NxEmOmMtexEbsdParser

from pynxtools.dataconverter.readers.em_om.utils.zip_ebsd_parser \
    import NxEmOmZipEbsdParser

from pynxtools.dataconverter.readers.em_om.utils.dream3d_ebsd_parser \
    import NxEmOmDreamThreedEbsdParser

from pynxtools.dataconverter.readers.em_om.utils.em_nexus_plots \
    import em_om_default_plot_generator


class EmOmReader(BaseReader):
    """Parse content from file formats of the electron-backscatter community.

    Specifically, electron microscopy
    towards a NXem_ebsd.nxdl-compliant NeXus file.
    """

    # pylint: disable=too-few-public-methods

    # Whitelist for the NXDLs that the reader supports and can process
    supported_nxdls = ["NXem_ebsd"]  # how to combine with "NXem"?

    # pylint: disable=duplicate-code
    def read(self,
             template: dict = None,
             file_paths: Tuple[str] = None,
             objects: Tuple[Any] = None) -> dict:
        """Read data from given file, return filled template dictionary em."""
        # pylint: disable=duplicate-code
        template.clear()

        # this em_om parser combines multiple sub-parsers
        # so we need the following input:
        # logical analysis which use case
        # data input from an ELN (using an ELN-agnostic) YAML representation
        # data input from technology partner files
        # functionalities for creating default plots

        entry_id = 1
        # if len(file_paths) != 2:
        #     print("Generation of example data not implemented yet...!")
        #     return {}

        print("Parse ELN and technology partner file(s)...")
        case = EmOmUseCaseSelector(file_paths)
        if case.is_valid is False:
            print("Such a combination of input-file(s, if any) is not supported !")
            return {}

        print("Parse (meta)data coming from an ELN...")
        if case.eln_parser_type == "generic":
            pattern_simulation = False
            if case.dat_parser_type == "zip":
                pattern_simulation = True
            eln = NxEmOmGenericElnSchemaParser(case.eln[0], entry_id, pattern_simulation)
            eln.parse(template)
        else:
            print("No interpretable ELN input found!")
            return {}

        print("Parse (numerical) data and metadata from technology partner files...")
        if case.dat_parser_type == "orix":
            orix_parser = NxEmOmOrixEbsdParser(case.dat[0], entry_id)
            # h5oina parser evaluating content and plotting with orix on the fly
            orix_parser.parse(template)
        elif case.dat_parser_type == "mtex":
            mtex_parser = NxEmOmMtexEbsdParser(case.dat[0], entry_id)
            # ebsd parser because concept suggested for MTex by M. Kühbach
            # would include different HDF5 dumps for different MTex classes
            mtex_parser.parse(template)
        elif case.dat_parser_type == "zip":
            zip_parser = NxEmOmZipEbsdParser(case.dat[0], entry_id)
            zip_parser.parse(template)
        elif case.dat_parser_type == "dream3d":
            dream_parser = NxEmOmDreamThreedEbsdParser(case.dat[0], entry_id)
            dream_parser.parse(template)
        # elif case.dat_parser_type == "kikuchipy":
        # elif case.dat_parser_type == "pyxem":
        # elif case.dat_parser_type == "score":
        # elif case.dat_parser_type == "qube":
        # elif case.dat_parser_type == "paradis":
        # elif case.dat_parser_type == "brinckmann":
        else:
            print("No input-file defined for technology partner data !")

        # at this point the default plots exist already
        # we only need to decorate the template to point to the mandatory ROI overview
        print("Create NeXus default plottable data...")
        em_om_default_plot_generator(template, 1)

        debugging = False
        if debugging is True:
            print("Reporting state of template before passing to HDF5 writing...")
            for keyword in template.keys():
                print(keyword)
                # print(type(template[keyword]))
                # print(template[keyword])

        print("Forward instantiated template to the NXS writer...")
        return template


# This has to be set to allow the convert script to use this reader.
READER = EmOmReader



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Load deployment-specific quantities."""

# pylint: disable=no-member

import flatdict as fd

import yaml

from pynxtools.dataconverter.readers.apm.map_concepts.apm_deployment_specifics_to_nx_map \
    import NxApmDeploymentSpecificInput

from pynxtools.dataconverter.readers.shared.map_concepts.mapping_functors \
    import apply_modifier, variadic_path_to_specific_path


class NxApmNomadOasisConfigurationParser:  # pylint: disable=too-few-public-methods
    """Parse deployment specific configuration."""

    def __init__(self, file_name: str, entry_id: int):
        print(f"Extracting data from deployment specific configuration file: {file_name}")
        if (file_name.rsplit('/', 1)[-1].endswith(".oasis.specific.yaml")
                or file_name.endswith(".oasis.specific.yml")) and entry_id > 0:
            self.entry_id = entry_id
            self.file_name = file_name
            with open(self.file_name, "r", encoding="utf-8") as stream:
                self.yml = fd.FlatDict(yaml.safe_load(stream), delimiter="/")
        else:
            self.entry_id = 1
            self.file_name = ""
            self.yml = {}

    def report(self, template: dict) -> dict:
        """Copy data from configuration applying mapping functors."""
        for nx_path, modifier in NxApmDeploymentSpecificInput.items():
            if nx_path not in ("IGNORE", "UNCLEAR"):
                trg = variadic_path_to_specific_path(nx_path, [self.entry_id, 1])
                res = apply_modifier(modifier, self.yml)
                if res is not None:
                    template[trg] = res
        return template



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Dict mapping values for a specific deployed config of NOMAD OASIS + ELN + apm reader."""

# pylint: disable=no-member,line-too-long

# currently by virtue of design NOMAD OASIS specific examples show how different tools and
# services can be specifically coupled and implemented so that they work together
# currently we assume that the ELN provides all those pieces of information to instantiate
# a NeXus data artifact which technology-partner-specific files or database blobs can not
# deliver. Effectively a reader uses the eln_data.yaml generic ELN output to fill in these
# missing pieces of information while typically heavy data (tensors etc) are translated
# and written from the technology-partner files
# for large application definitions this can lead to a practical inconvenience:
# the ELN that has to be exposed to the user is complex and has many fields to fill in
# just to assure that all information are included in the ELN output and thus consumable
# by the dataconverter
# taking the perspective of a specific lab where a specific version of an ELN provided by
# or running in addition to NOMAD OASIS is used many pieces of information might not change
# or administrators do not wish to expose this via the end user ELN in an effort to reduce
# the complexity for end users and make entering of repetitiv information obsolete

# this is the scenario for which deployment_specific mapping shines
# parsing of deployment specific details in the apm reader is currently implemented
# such that it executes after reading generic ELN data (eventually available entries)
# in the template get overwritten

from pynxtools.dataconverter.readers.apm.utils.apm_versioning \
    import NX_APM_ADEF_NAME, NX_APM_ADEF_VERSION, NX_APM_EXEC_NAME, NX_APM_EXEC_VERSION


NxApmDeploymentSpecificInput \
    = {"/ENTRY[entry*]/@version": f"{NX_APM_ADEF_VERSION}",
       "/ENTRY[entry*]/definition": f"{NX_APM_ADEF_NAME}",
       "/ENTRY[entry*]/PROGRAM[program1]/program": f"{NX_APM_EXEC_NAME}",
       "/ENTRY[entry*]/PROGRAM[program1]/program/@version": f"{NX_APM_EXEC_VERSION}",
       "/ENTRY[entry*]/atom_probe/location": {"fun": "load_from", "terms": "location"}}
